from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="CollectionsItemsInitiateMultipartUploadBody")


@_attrs_define
class CollectionsItemsInitiateMultipartUploadBody:
    """CollectionsItemsInitiateMultipartUploadBody model

    Attributes:
        size (int):
        datastore (Union[Unset, str]):  Default: 'delfini+datastore://default'.
        num_parts (Union[Unset, int]):
    """

    size: int
    datastore: Union[Unset, str] = "delfini+datastore://default"
    num_parts: Union[Unset, int] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        size = self.size
        datastore = self.datastore
        num_parts = self.num_parts

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "size": size,
            }
        )
        if datastore is not UNSET:
            field_dict["datastore"] = datastore
        if num_parts is not UNSET:
            field_dict["numParts"] = num_parts

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`CollectionsItemsInitiateMultipartUploadBody` from a dict"""
        d = src_dict.copy()
        size = d.pop("size")

        datastore = d.pop("datastore", UNSET)

        num_parts = d.pop("numParts", UNSET)

        collections_items_initiate_multipart_upload_body = cls(
            size=size,
            datastore=datastore,
            num_parts=num_parts,
        )

        return collections_items_initiate_multipart_upload_body



"""File-like (stream) read and write on items"""
import io
import os
import threading
from collections.abc import Iterable
from collections.abc import Iterator
from types import TracebackType
from typing import BinaryIO
from typing import Literal
from typing import Optional
from typing import overload
from typing import TextIO
from typing import TYPE_CHECKING
from typing import Union

if TYPE_CHECKING:
    from typing_extensions import Buffer

from .delfini_core import AuthenticatedClient as CoreAuthClient
from .delfini_core.api.items import collections_items_create_item
from .delfini_core.api.items import collections_items_get_item_data
from .delfini_core.api.items import collections_items_put_item_data
from .delfini_core.models import CollectionsItemsCreateItemFilesFormStyleUpload
from .delfini_core.models import CollectionsItemsCreateItemFilesFormStyleUploadMetadata
from .delfini_core.models import CollectionsItemsCreateItemJsonNewItemRequest
from .delfini_core.models import Item
from .delfini_core.models import ItemColumn
from .delfini_core.models import ItemParser
from .delfini_core.models import ItemStorage
from .delfini_core.models import ItemType
from .delfini_core.types import File
from .delfini_core.types import UNSET

ItemReadModes = Literal["r", "rb"]
ItemWriteModes = Literal["w", "wb"]
ItemBinaryModes = Literal["rb", "wb"]
ItemTextModes = Literal["r", "w"]
ItemIOModes = Literal[ItemReadModes, ItemWriteModes]


# When doing stream IO on items (reading and writing bytes), one of
# the challenging goals is to try to read and write directly from the
# HTTP stream; that is, to avoid buffering the data into a secondary
# stream that the user interacts with. This is important because it
# minimizes memory usage and latency from copying large amounts of
# data.
#
# When reading items, direct access to the item's stream is
# straightforward, since the httpx client used by delfini_core returns
# a `payload` that can be treated directly as a stream. The only
# detail is to support both binary and text modes of interaction,
# however that is easily handled with io.TextIOWrapper.
#
# However, it's much harder when writing the item. The httpx client
# can accept a stream when submitting the item data, but the user also
# needs a stream to write to. Rather than buffering the user's write
# content in memory via something like io.BytesIO, we do it the UNIX
# way -- we create a _pipe_, give the httpx client the read end of the
# pipe, and give the user the write end. Then, the API call is run in
# a thread so that the pipe does not fill up and block. In order to
# manage the thread, the user should be sure to either `close()` the
# file handle when finished, or else use it in a context manager
# block.


@overload
def read_item(
    mode: Literal["r"],
    collection_id: str,
    version_id: str,
    item_id: str,
    client: CoreAuthClient,
) -> TextIO:
    ...


@overload
def read_item(
    mode: Literal["rb"],
    collection_id: str,
    version_id: str,
    item_id: str,
    client: CoreAuthClient,
) -> BinaryIO:
    ...


def read_item(
    mode: ItemReadModes,
    collection_id: str,
    version_id: str,
    item_id: str,
    client: CoreAuthClient,
) -> Union[BinaryIO, TextIO]:
    """Read the contents (binary or text mode) of an item.

    This is typically called by the high-level client libraries in
    :py:mod:`pydelfini.collections`.

    """
    response = collections_items_get_item_data.sync(
        collection_id=collection_id,
        version_id=version_id,
        item_id=item_id,
        client=client,
    )
    if mode == "rb":
        return response.payload
    else:
        return io.TextIOWrapper(response.payload)


class DelfiniItemWriterCore:
    def __init__(
        self, collection_id: str, version_id: str, item_id: str, client: CoreAuthClient
    ) -> None:
        self.collection_id = collection_id
        self.version_id = version_id
        self.item_id = item_id
        self._client = client
        self._exc: Optional[BaseException] = None

        # Create a pipe - read side is for the API client, write side
        # is for the user
        read_side, write_side = os.pipe()
        self._binary_stream = open(write_side, "wb")
        read_stream = open(read_side, "rb")

        # Patch out the fileno method on the read stream so that httpx
        # doesn't think this is a real file and lets us treat it as a
        # streaming body.
        read_stream.fileno = self.fileno  # type: ignore[method-assign]

        # The API call needs to be in a thread so that it can consume
        # the read side of the pipe without the write side blocking
        self._thread = threading.Thread(
            target=self._call_api, kwargs={"stream": read_stream}
        )
        self._thread.start()

    def _call_api(self, stream: BinaryIO) -> None:
        try:
            collections_items_put_item_data.sync(
                self.collection_id,
                self.version_id,
                self.item_id,
                client=self._client,
                body=File(payload=stream),
            )
        except BaseException as e:
            self._exc = e

    def __exit__(
        self,
        typ: Optional[type[BaseException]],
        val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        self._binary_stream.close()
        self._thread.join()

        if self._exc:
            raise self._exc

    def close(self) -> None:
        """Close and flush the stream, completing the file write."""
        self.__exit__(None, None, None)

    def __del__(self) -> None:
        self.close()

    def fileno(self) -> int:
        """N/A - not a real file

        Raises:
            OSError: always

        """
        raise OSError("not a real file")

    def flush(self) -> None:
        self._binary_stream.flush()

    def isatty(self) -> bool:
        """Always False"""
        return False

    def readable(self) -> bool:
        """Always False"""
        return False

    def writable(self) -> bool:
        """Always True"""
        return True

    def seekable(self) -> bool:
        """Always False"""
        return False

    def seek(self, offset: int, whence: int = 0) -> int:
        """N/A - not a real file

        Raises:
            OSError: always

        """
        raise OSError("not a real file")

    def tell(self) -> int:
        return self._binary_stream.tell()

    def truncate(self, size: Optional[int] = None) -> int:
        """N/A - not a real file

        Raises:
            OSError: always

        """
        raise OSError("not a real file")


class DelfiniItemWriterBinary(DelfiniItemWriterCore, BinaryIO):
    r"""File-like interface for writing binary data to items.

    It is important to either use this in a context manager or else
    call the :py:func:`close` method when writing is complete,
    otherwise the full contents of the file may not be written.

    Suggested usage::

        with item.open('wb') as fp:
            fp.write(b'my item content, as much as I have\n')

    """

    def __enter__(self) -> BinaryIO:
        return self._binary_stream

    def __iter__(self) -> Iterator[bytes]:
        raise OSError("cannot read")

    def __next__(self) -> bytes:
        raise OSError("cannot read")

    def read(self, n: int = -1) -> bytes:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def readline(self, limit: int = -1) -> bytes:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def readlines(self, hint: int = -1) -> list[bytes]:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def write(self, s: Union[bytes, "Buffer"]) -> int:
        """Write to the stream.

        Args:
            s: content to write

        Returns:
            number of bytes written
        """
        return self._binary_stream.write(s)

    def writelines(self, lines: Union[Iterable[bytes], Iterable["Buffer"]]) -> None:
        """Write lines to the stream.

        Args:
            lines: iterable of lines (bytes or Buffer) to write

        """
        return self._binary_stream.writelines(lines)


class DelfiniItemWriterText(DelfiniItemWriterCore, TextIO):
    r"""File-like interface for writing text data to items.

    It is important to either use this in a context manager or else
    call the :py:func:`close` method when writing is complete,
    otherwise the full contents of the file may not be written.

    Suggested usage::

        with item.open('w') as fp:
            fp.write('my item content, as much as I have\n')

    """

    def __init__(
        self, collection_id: str, version_id: str, item_id: str, client: CoreAuthClient
    ) -> None:
        super().__init__(collection_id, version_id, item_id, client)
        self._text_stream = io.TextIOWrapper(self._binary_stream)

    def __enter__(self) -> TextIO:
        return self._text_stream

    def __exit__(
        self,
        typ: Optional[type[BaseException]],
        val: Optional[BaseException],
        tb: Optional[TracebackType],
    ) -> None:
        self._text_stream.close()
        super().__exit__(typ, val, tb)

    def __iter__(self) -> Iterator[str]:
        raise OSError("cannot read")

    def __next__(self) -> str:
        raise OSError("cannot read")

    def read(self, n: int = -1) -> str:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def readline(self, limit: int = -1) -> str:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def readlines(self, hint: int = -1) -> list[str]:
        """N/A - cannot read

        Raises:
            OSError: always

        """
        raise OSError("cannot read")

    def write(self, s: str) -> int:
        """Write to the stream.

        Args:
            s: content to write

        Returns:
            number of characters written
        """
        return self._text_stream.write(s)

    def writelines(self, lines: Iterable[str]) -> None:
        """Write lines to the stream.

        Args:
            lines: iterable of lines (str) to write

        """
        return self._text_stream.writelines(lines)


class DelfiniItemCreatorCore(DelfiniItemWriterCore):
    def __init__(
        self,
        collection_id: str,
        version_id: str,
        item_name: str,
        client: CoreAuthClient,
        folder_id: str = "ROOT",
        type: Literal["file", "dataview", "dictionary"] = "file",
        columns: Optional[list[ItemColumn]] = None,
        metadata: Optional[dict[str, str]] = None,
        parser: Optional[ItemParser] = None,
        storage: Optional[ItemStorage] = None,
    ) -> None:
        self.item_name = item_name
        self.item_folder_id = folder_id
        self.item_type = type
        self.item_columns = columns
        self.item_metadata = metadata
        self.item_parser = parser
        self.item_storage = storage
        super().__init__(collection_id, version_id, "_pending_close", client)

    def _call_api(self, stream: BinaryIO) -> None:
        try:
            request = CollectionsItemsCreateItemFilesFormStyleUpload(
                content=File(payload=stream, file_name=self.item_name),
                folder_id=self.item_folder_id,
                type=ItemType[self.item_type.upper()],
                columns=self.item_columns if self.item_columns else UNSET,
                metadata=(
                    CollectionsItemsCreateItemFilesFormStyleUploadMetadata.from_dict(
                        self.item_metadata
                    )
                    if self.item_metadata
                    else UNSET
                ),
                parser=self.item_parser if self.item_parser else UNSET,
                storage=self.item_storage if self.item_storage else UNSET,
            )
            item = collections_items_create_item.sync(
                self.collection_id,
                self.version_id,
                body=request,
                client=self._client,
            )
            self.item_id = item.id
        except BaseException as e:
            print("caught exception:", e)
            self._exc = e


class DelfiniItemCreatorBinary(DelfiniItemCreatorCore, DelfiniItemWriterBinary):
    r"""File-like interface for creating a new item in binary mode.

    It is important to either use this in a context manager or else
    call the :py:func:`close` method when writing is complete,
    otherwise the full contents of the file may not be written.

    Suggested usage::

        with collection.open('new-item', 'wb') as fp:
            fp.write(b'my item content, as much as I have\n')

    """
    pass


class DelfiniItemCreatorText(DelfiniItemCreatorCore, DelfiniItemWriterText):
    r"""File-like interface for creating a new item in text mode.

    It is important to either use this in a context manager or else
    call the :py:func:`close` method when writing is complete,
    otherwise the full contents of the file may not be written.

    Suggested usage::

        with collection.open('new-item', 'w') as fp:
            fp.write('my item content, as much as I have\n')

    """
    pass


def new_empty_item(
    collection_id: str,
    version_id: str,
    name: str,
    type: Literal["file", "folder", "dataview", "dictionary"],
    client: CoreAuthClient,
    within_folder_id: str = "ROOT",
) -> Item:
    request = CollectionsItemsCreateItemJsonNewItemRequest(
        name=name,
        type=ItemType[type.upper()],
        folder_id=within_folder_id,
    )
    item = collections_items_create_item.sync(
        collection_id, version_id, body=request, client=client
    )
    return item


def new_folder(
    collection_id: str,
    version_id: str,
    name: str,
    client: CoreAuthClient,
    within_folder_id: str = "ROOT",
) -> Item:
    return new_empty_item(
        collection_id,
        version_id,
        name,
        "folder",
        client,
        within_folder_id=within_folder_id,
    )



import os.path
import shutil

from pydelfini.delfini_core.api.collections import collections_create_collection
from pydelfini.delfini_core.api.collections import collections_delete_collection_version
from pydelfini.delfini_core.api.collections import collections_get_collections
from pydelfini.delfini_core.models import NewCollection

from .base_commands import BaseCommands


class CollectionCommands(BaseCommands):
    """Operations on collections"""

    def list(self) -> None:
        """List all visible collections."""
        collections = collections_get_collections.sync(client=self.core)

        self._output(collections.to_dict())

    @BaseCommands._with_arg("name", help="Collection name, required")
    @BaseCommands._with_arg(
        "-d", "--description", help="Collection description", default=""
    )
    def new(self) -> None:
        """Create a new collection."""
        collection = collections_create_collection.sync(
            body=NewCollection(
                name=self.args.name,
                description=self.args.description,
            ),
            client=self.core,
        )

        self._output(collection.to_dict())

    @BaseCommands._with_arg("collection_id", help="Collection ID")
    @BaseCommands._with_arg(
        "-v", "--version_id", help="Collection version ID", default="LIVE"
    )
    def delete(self) -> None:
        """Delete a collection."""
        rv = collections_delete_collection_version.sync(
            collection_id=self.args.collection_id,
            version_id=self.args.version_id,
            client=self.core,
        )
        assert rv

    @BaseCommands._with_arg("collection_name", help="Collection name")
    @BaseCommands._with_arg("filename", help="File to upload")
    def upload(self) -> None:
        """Upload a file to a collection."""
        collection = self.client.get_collection_by_name(self.args.collection_name)

        with open(self.args.filename, "rb") as src, collection.open(
            os.path.basename(self.args.filename), "wb"
        ) as dest:
            shutil.copyfileobj(src, dest)

    @BaseCommands._with_arg("collection_name", help="Collection name")
    @BaseCommands._with_arg("filename", help="File to download")
    def download(self) -> None:
        """Download an item from a collection."""
        collection = self.client.get_collection_by_name(self.args.collection_name)

        with collection.open(self.args.filename, "rb") as src, open(
            os.path.basename(self.args.filename), "wb"
        ) as dest:
            shutil.copyfileobj(src, dest)



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define


T = TypeVar("T", bound="TermsetAdditionalPropertyItem")


@_attrs_define
class TermsetAdditionalPropertyItem:
    """TermsetAdditionalPropertyItem model

    Attributes:
        docs (int):
        term (str):
    """

    docs: int
    term: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        docs = self.docs
        term = self.term

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "docs": docs,
                "term": term,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`TermsetAdditionalPropertyItem` from a dict"""
        d = src_dict.copy()
        docs = d.pop("docs")

        term = d.pop("term")

        termset_additional_property_item = cls(
            docs=docs,
            term=term,
        )

        return termset_additional_property_item



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.highlights import Highlights


T = TypeVar("T", bound="SearchCollectionsResponseHitsItemDataElementsItem")


@_attrs_define
class SearchCollectionsResponseHitsItemDataElementsItem:
    """SearchCollectionsResponseHitsItemDataElementsItem model

    Attributes:
        highlights (Highlights):
        id (str):
        url (str):
        version (str):
    """

    highlights: "Highlights"
    id: str
    url: str
    version: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        highlights = self.highlights.to_dict()
        id = self.id
        url = self.url
        version = self.version

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "highlights": highlights,
                "id": id,
                "url": url,
                "version": version,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SearchCollectionsResponseHitsItemDataElementsItem` from a dict"""
        d = src_dict.copy()
        highlights = Highlights.from_dict(d.pop("highlights"))

        id = d.pop("id")

        url = d.pop("url")

        version = d.pop("version")

        search_collections_response_hits_item_data_elements_item = cls(
            highlights=highlights,
            id=id,
            url=url,
            version=version,
        )

        return search_collections_response_hits_item_data_elements_item



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.collection_authorization_grant import CollectionAuthorizationGrant
from ..models.collection_authorization_inherited import CollectionAuthorizationInherited
from ..models.collection_authorization_permit import CollectionAuthorizationPermit
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="CollectionAuthorization")


@_attrs_define
class CollectionAuthorization:
    """CollectionAuthorization model

    Attributes:
        controlled_access_groups (List[str]):
        myself (CollectionAuthorizationGrant):
        public (CollectionAuthorizationGrant):
        inherited (Union[Unset, CollectionAuthorizationInherited]):
        permit (Union[Unset, CollectionAuthorizationPermit]):
    """

    controlled_access_groups: List[str]
    myself: "CollectionAuthorizationGrant"
    public: "CollectionAuthorizationGrant"
    inherited: Union[Unset, "CollectionAuthorizationInherited"] = UNSET
    permit: Union[Unset, "CollectionAuthorizationPermit"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        controlled_access_groups = self.controlled_access_groups

        myself = self.myself.to_dict()
        public = self.public.to_dict()
        inherited: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.inherited, Unset):
            inherited = self.inherited.to_dict()
        permit: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.permit, Unset):
            permit = self.permit.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "controlledAccessGroups": controlled_access_groups,
                "myself": myself,
                "public": public,
            }
        )
        if inherited is not UNSET:
            field_dict["inherited"] = inherited
        if permit is not UNSET:
            field_dict["permit"] = permit

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`CollectionAuthorization` from a dict"""
        d = src_dict.copy()
        controlled_access_groups = cast(List[str], d.pop("controlledAccessGroups"))

        myself = CollectionAuthorizationGrant.from_dict(d.pop("myself"))

        public = CollectionAuthorizationGrant.from_dict(d.pop("public"))

        _inherited = d.pop("inherited", UNSET)
        inherited: Union[Unset, CollectionAuthorizationInherited]
        if isinstance(_inherited, Unset):
            inherited = UNSET
        else:
            inherited = CollectionAuthorizationInherited.from_dict(_inherited)

        _permit = d.pop("permit", UNSET)
        permit: Union[Unset, CollectionAuthorizationPermit]
        if isinstance(_permit, Unset):
            permit = UNSET
        else:
            permit = CollectionAuthorizationPermit.from_dict(_permit)

        collection_authorization = cls(
            controlled_access_groups=controlled_access_groups,
            myself=myself,
            public=public,
            inherited=inherited,
            permit=permit,
        )

        return collection_authorization



"""Resolve Data Element URL"""
from http import HTTPStatus
from typing import Any
from typing import Dict
from typing import Union

import httpx

from ... import errors
from ...client import AuthenticatedClient
from ...client import Client
from ...models.data_element import DataElement
from ...models.server_error import ServerError
from ...types import Response
from ...types import UNSET
from ...types import Unset


def _get_kwargs(
    *,
    url_query: str,
    collection_id: Union[Unset, str] = UNSET,
    version_id: Union[Unset, str] = UNSET,
) -> Dict[str, Any]:
    params: Dict[str, Any] = {}

    params["url"] = url_query

    params["collection_id"] = collection_id

    params["version_id"] = version_id

    params = {k: v for k, v in params.items() if v is not UNSET and v is not None}

    _kwargs: Dict[str, Any] = {
        "method": "get",
        "url": "/resolve",
        "params": params,
    }

    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[DataElement, ServerError]:
    if response.status_code == HTTPStatus.OK:
        response_200 = DataElement.from_dict(response.json())

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ServerError.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = ServerError.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.NOT_FOUND:
        response_404 = ServerError.from_dict(response.json())

        return response_404
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = ServerError.from_dict(response.json())

        return response_500

    raise errors.UnexpectedStatus(response.status_code, response.content)


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[DataElement, ServerError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    url_query: str,
    collection_id: Union[Unset, str] = UNSET,
    version_id: Union[Unset, str] = UNSET,
) -> Response[Union[DataElement, ServerError]]:
    """Resolve Data Element URL

    Retrieve a data element from a data element URL

    Args:
        url_query (str):
        collection_id (Union[Unset, str]):
        version_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DataElement, ServerError]]
    """

    kwargs = _get_kwargs(
        url_query=url_query,
        collection_id=collection_id,
        version_id=version_id,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: Union[AuthenticatedClient, Client],
    url_query: str,
    collection_id: Union[Unset, str] = UNSET,
    version_id: Union[Unset, str] = UNSET,
) -> Union[DataElement]:
    """Resolve Data Element URL

    Retrieve a data element from a data element URL

    Args:
        url_query (str):
        collection_id (Union[Unset, str]):
        version_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DataElement]
    """

    response = sync_detailed(
        client=client,
        url_query=url_query,
        collection_id=collection_id,
        version_id=version_id,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed


async def asyncio_detailed(
    *,
    client: Union[AuthenticatedClient, Client],
    url_query: str,
    collection_id: Union[Unset, str] = UNSET,
    version_id: Union[Unset, str] = UNSET,
) -> Response[Union[DataElement, ServerError]]:
    """Resolve Data Element URL

    Retrieve a data element from a data element URL

    Args:
        url_query (str):
        collection_id (Union[Unset, str]):
        version_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[DataElement, ServerError]]
    """

    kwargs = _get_kwargs(
        url_query=url_query,
        collection_id=collection_id,
        version_id=version_id,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: Union[AuthenticatedClient, Client],
    url_query: str,
    collection_id: Union[Unset, str] = UNSET,
    version_id: Union[Unset, str] = UNSET,
) -> Union[DataElement]:
    """Resolve Data Element URL

    Retrieve a data element from a data element URL

    Args:
        url_query (str):
        collection_id (Union[Unset, str]):
        version_id (Union[Unset, str]):

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[DataElement]
    """

    response = await asyncio_detailed(
        client=client,
        url_query=url_query,
        collection_id=collection_id,
        version_id=version_id,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed



"""Interactions with collections, items, and data tables"""
import abc
import os.path
from collections.abc import Iterator
from datetime import datetime
from typing import Any
from typing import BinaryIO
from typing import cast
from typing import Literal
from typing import Optional
from typing import overload
from typing import TextIO
from typing import TYPE_CHECKING
from typing import Union

import pandas as pd
import pyarrow as pa
from tqdm.auto import tqdm

from . import item_io
from .delfini_core import AuthenticatedClient as CoreAuthClient
from .delfini_core import Paginator
from .delfini_core.api.collections import collections_delete_collection_version
from .delfini_core.api.items import collections_items_list_items
from .delfini_core.api.tables import collections_tables_get_table_data
from .delfini_core.models import Collection
from .delfini_core.models import CollectionsItemsListItemsResponse200
from .delfini_core.models import Item
from .delfini_core.models import ItemParser
from .delfini_core.models import ItemType
from .delfini_core.models import TableData
from .delfini_core.types import UNSET


def pyarrow_schema_from_json_schema(js: dict[str, Any]) -> pa.Schema:
    """Build a PyArrow schema from a JSON schema.

    Args:
        js: JSON schema as dict

    """
    schema_parts = []
    for field, definition in js["properties"].items():
        field_s: dict[str, Any] = {"name": field}
        js_type = definition.get("type")
        js_format = definition.get("format")
        if js_type == "integer":
            field_s["type"] = pa.int64()
        elif js_type == "number":
            field_s["type"] = pa.float64()
        elif js_type == "string":
            if js_format == "date":
                field_s["type"] = pa.date32()
            elif js_format == "time":
                field_s["type"] = pa.time32()
            elif js_format == "datetime":
                field_s["type"] = pa.timestamp("s")
            elif js_format == "duration":
                field_s["type"] = pa.duration("s")
            else:
                field_s["type"] = pa.string()
        elif js_type == "boolean":
            field_s["type"] = pa.bool_()

        if "type" not in field_s:
            raise Exception(f"not yet supporting JSON schema: {definition}")

        # TODO: add data element ref
        field_s["metadata"] = {"dataElement": "..."}

        schema_parts.append(pa.field(**field_s))

    return pa.schema(schema_parts)


class FolderMixin(abc.ABC):
    """Folder interaction methods as a mixin class.

    This class is used by :py:class:`DelfiniCollection` and
    :py:class:`DelfiniFolder`.

    """

    def __init__(
        self,
        collection: "DelfiniCollection",
        full_path: Optional[str],
        core: CoreAuthClient,
    ) -> None:
        self._top = collection
        self._path = full_path
        self._client = core

    @property
    @abc.abstractmethod
    def id(self) -> str:
        raise NotImplementedError

    def _join_path(self, append: str) -> str:
        if self._path is None:
            return append
        return os.path.join(self._path, append)

    def folder(self, folder_name: str) -> "DelfiniFolder":
        """Open the named folder.

        Args:
            folder_name (str): the name of the folder to open
        """
        model = collections_items_list_items.sync(
            client=self._client,
            collection_id=self._top.id,
            version_id=self._top.version_id,
            in_path=UNSET if self._path is None else self._path,
            in_folder="ROOT" if self._path is None else UNSET,
            name=folder_name,
            type=[ItemType.FOLDER],
        )
        if len(model.items) != 1:
            raise Exception(f"folder not found: {folder_name}")
        return DelfiniFolder(self._top, self, model.items[0], self._client)

    def new_folder(self, folder_name: str) -> "DelfiniFolder":
        """Create a new folder.

        Args:
            folder_name (str): the name of the folder to create
        """
        within_folder_id = "ROOT" if self._top.id == self.id else self.id
        folder = item_io.new_folder(
            self._top.id,
            self._top.version_id,
            folder_name,
            self._client,
            within_folder_id=within_folder_id,
        )
        return DelfiniFolder(self._top, self, folder, self._client)

    def new_empty_file(self, file_name: str) -> "DelfiniItem":
        """Create a new, empty file.

        Args:
            file_name (str): the name of the file to create
        """
        within_folder_id = "ROOT" if self._top.id == self.id else self.id
        item = item_io.new_empty_item(
            self._top.id,
            self._top.version_id,
            file_name,
            "file",
            self._client,
            within_folder_id=within_folder_id,
        )
        return DelfiniItem(self._top, self, item, self._client)

    def __getitem__(self, item_path: str) -> "DelfiniItem":
        """Retrieve an item in this folder.

        Relative paths are supported, such as ``folder_a/folder_b/item_c``.
        """
        if "/" in item_path:
            folder_name, path_remain = item_path.split("/", 1)
            return self.folder(folder_name)[path_remain]

        model = collections_items_list_items.sync(
            client=self._client,
            collection_id=self._top.id,
            version_id=self._top.version_id,
            in_path=UNSET if self._path is None else self._path,
            in_folder="ROOT" if self._path is None else UNSET,
            name=item_path,
        )
        if len(model.items) != 1:
            if self._path is None:
                msg = f"in top level, item not found: {item_path}"
            else:
                msg = f"in folder {self._path}, item not found: {item_path}"
            raise Exception(msg)
        return DelfiniItem(self._top, self, model.items[0], self._client)

    def __iter__(self) -> Iterator["DelfiniItem"]:
        """Iterate over all items in this folder."""
        paginator = Paginator[CollectionsItemsListItemsResponse200](
            collections_items_list_items,
            self._client,
        )
        kwargs = {"collection_id": self._top.id, "version_id": self._top.version_id}
        if self._path is None:
            kwargs["in_folder"] = "ROOT"
        else:
            kwargs["in_path"] = self._path

        for item_page in paginator.paginate(**kwargs):
            for item in item_page.items:
                yield DelfiniItem(self._top, self, item, self._client)

    def walk(self) -> Iterator["DelfiniItem"]:
        """Iterate over all items in this folder, and all subfolders."""
        for item in self:
            yield item
            if isinstance(item, DelfiniFolder):
                yield from item

    def get_table(self, item_path: str) -> pd.DataFrame:
        """Retrieve the tabular contents of an item as a :py:class:`pd.DataFrame`.

        Args:
            item_path:
                The path to the item relative to the current folder.
                Accepts slashes ('/') for items in subfolders.

        """
        return self[item_path].table()

    def write_table(
        self,
        item_path: str,
        dataframe: pd.DataFrame,
        format: Literal["csv", "parquet"] = "csv",
        **kwargs: Any,
    ) -> None:
        """Write a :py:class:`pd.DataFrame` to the named item.

        Args:
            item_path:
                The path to the item relative to the current folder.
                Accepts slashes ('/') for items in subfolders.

            dataframe: The dataframe to be written.

            format: One of the supported formats (``csv`` or ``parquet``).

            **kwargs:
                Any other arguments to be passed to the Pandas export
                function. See the documentation for
                :py:func:`pd.DataFrame.to_csv` or
                :py:func:`pd.DataFrame.to_parquet` for valid
                arguments.

        """
        metadata = {
            "Content-Type": {
                "csv": "text/csv",
                "parquet": "application/vnd.apache.parquet",
            }[format]
        }
        with self._top.open(item_path, "wb", parser=format, metadata=metadata) as fp:
            if format == "csv":
                kwargs.setdefault("index", False)
                dataframe.to_csv(fp, mode="wb", **kwargs)
            elif format == "parquet":
                dataframe.to_parquet(fp, **kwargs)


if TYPE_CHECKING:
    ColumnsType = Union[pd.Series[type[object]], pa.Schema]
else:
    ColumnsType = Union[pd.Series, pa.Schema]


class DelfiniCollection(FolderMixin):
    """Represents a collection on a Delfini instance.

    Typically created by one of the collection methods in
    :py:class:`pydelfini.client.DelfiniClient`.

    In addition to collection-specific attributes and methods, this
    class also behaves as a folder (representing the top level of the
    collection's folder structure). See :py:class:`FolderMixin` for
    those methods.

    """

    def __init__(self, model: Collection, core: CoreAuthClient) -> None:
        self._model = model
        FolderMixin.__init__(self, self, None, core)

    @property
    def name(self) -> str:
        """Collection name"""
        return self._model.name

    @property
    def id(self) -> str:
        """Collection internal ID"""
        return self._model.id

    @property
    def version_id(self) -> str:
        """Current version ID"""
        return self._model.version_id

    @property
    def description(self) -> str:
        """Collection text description"""
        return self._model.description

    @property
    def created_on(self) -> datetime:
        """Datetime when collection was created"""
        return self._model.created_on

    def __repr__(self) -> str:
        return (
            f"<DelfiniCollection: name={self.name} version={self.version_id}"
            f" id={self.id}>"
        )

    def delete_collection(self) -> None:
        """Deletes the entire collection, permanently.

        You cannot delete the LIVE version of a collection if there
        exist other versions of the collection.

        """
        collections_delete_collection_version.sync(
            self.id, self.version_id, client=self._client
        )

    @overload
    def open(self, item_path: str, mode: Literal["r"]) -> TextIO:
        ...

    @overload
    def open(self, item_path: str, mode: Literal["rb"]) -> BinaryIO:
        ...

    @overload
    def open(
        self,
        item_path: str,
        mode: Literal["w"],
        *,
        parser: Optional[str] = None,
        columns: Optional[ColumnsType] = None,
        metadata: Optional[dict[str, str]] = None,
    ) -> TextIO:
        ...

    @overload
    def open(
        self,
        item_path: str,
        mode: Literal["wb"],
        *,
        parser: Optional[str] = None,
        columns: Optional[ColumnsType] = None,
        metadata: Optional[dict[str, str]] = None,
    ) -> BinaryIO:
        ...

    def open(
        self,
        item_path: str,
        mode: item_io.ItemIOModes = "r",
        *,
        parser: Optional[str] = None,
        columns: Optional[ColumnsType] = None,
        metadata: Optional[dict[str, str]] = None,
    ) -> Union[BinaryIO, TextIO]:
        r"""Open an item for reading or writing.

        When writing an item, it is important to use the returned
        file-like object either in a context manager or else call its
        :py:func:`close` method when writing is complete, otherwise
        the full contents of the file may not be written.

        Suggested usage::

            with collection.open('new-item', 'wb') as fp:
                fp.write(b'my item content, as much as I have\n')

        Args:
            item_path:
                The path to the item relative to the top level of the
                collection. Accepts slashes ('/') for items located in
                subfolders.

            mode:
                One of the following values, depending on the desired
                interaction mode:

                * ``r`` - reading, text mode
                * ``rb`` - reading, binary mode
                * ``w`` - writing, text mode
                * ``wb`` - writing, binary mode

            parser:
                If writing, optionally specify the item parser type.
                Typical values for this would be ``csv`` or
                ``parquet``.

            columns:
                If writing, optionally specify the item column schema.
                Requires ``parser`` to be set.

            metadata:
                If writing, optionally specify the item metadata.
                Typical values for this would be something like
                ``{"content-type": "text-csv"}``.

        Returns:
            A file-like interface, either :py:class:`BinaryIO` or
            :py:class:`TextIO` depending on whether the file was to be
            opened in binary or text mode. When writing, the return
            will be an instance of
            :py:class:`.item_io.DelfiniItemCreatorBinary` or
            :py:class:`.item_io.DelfiniItemCreatorText`.

        """
        # when reading, the item needs to exist, so this is the quickest way
        if mode in ("r", "rb"):
            return self[item_path].open(mode)

        elif mode in ("w", "wb"):
            _columns = None  # TODO: convert column schema
            _parser = ItemParser(name=parser) if parser else None
            _metadata = metadata

            if "/" in item_path:
                folder_path, item_name = item_path.rsplit("/", 1)
                folder_id = self[folder_path].id
            else:
                item_name = item_path
                folder_id = "ROOT"

            if mode == "w":
                return item_io.DelfiniItemCreatorText(
                    self.id,
                    self.version_id,
                    item_name,
                    self._client,
                    folder_id=folder_id,
                    type="file",
                    columns=_columns,
                    parser=_parser,
                    metadata=_metadata,
                )

            elif mode == "wb":
                return item_io.DelfiniItemCreatorBinary(
                    self.id,
                    self.version_id,
                    item_name,
                    self._client,
                    folder_id=folder_id,
                    type="file",
                    columns=_columns,
                    parser=_parser,
                    metadata=_metadata,
                )

        raise Exception("mode not found")


class DelfiniItem:
    """Represents an item within a collection."""

    #: Collection that contains this item
    collection: DelfiniCollection

    #: Folder that contains this item
    in_folder: FolderMixin

    def __new__(
        cls,
        collection: DelfiniCollection,
        in_folder: FolderMixin,
        model: Item,
        core: CoreAuthClient,
    ) -> "DelfiniItem":
        if model.type == ItemType.FOLDER and cls is DelfiniItem:
            # Automatically create an instance of DelfiniFolder when
            # the item is a folder
            return DelfiniFolder.__new__(
                DelfiniFolder, collection, in_folder, model, core
            )
        return super().__new__(cls)

    def __init__(
        self,
        collection: DelfiniCollection,
        in_folder: FolderMixin,
        model: Item,
        core: CoreAuthClient,
    ) -> None:
        self.collection = collection
        self.in_folder = in_folder
        self._model = model
        self._client = core

    @property
    def name(self) -> str:
        """Item name"""
        return self._model.name

    @property
    def path(self) -> str:
        """Item fully qualified path"""
        return self.in_folder._join_path(self.name)

    @property
    def id(self) -> str:
        """Item internal ID"""
        return self._model.id

    @property
    def type(self) -> ItemType:
        """Item type (``file``, ``folder``, ``dataview``, etc.)"""
        return self._model.type

    @property
    def created_on(self) -> datetime:
        """Datetime item was created"""
        return self._model.created_on

    def __repr__(self) -> str:
        return (
            f"<DelfiniItem: name={self.name} path={self.path} type={self.type}"
            f" id={self.id}>"
        )

    @overload
    def open(self, mode: item_io.ItemBinaryModes) -> BinaryIO:
        ...

    @overload
    def open(self, mode: item_io.ItemTextModes) -> TextIO:
        ...

    def open(self, mode: item_io.ItemIOModes = "r") -> Union[BinaryIO, TextIO]:
        r"""Open this item for reading or writing.

        When writing an item, it is important to use the returned
        file-like object either in a context manager or else call its
        :py:func:`close` method when writing is complete, otherwise
        the full contents of the file may not be written.

        Suggested usage::

            with item.open('wb') as fp:
                fp.write(b'my item content, as much as I have\n')

        Args:
            mode:
                One of the following values, depending on the desired
                interaction mode:

                * ``r`` - reading, text mode
                * ``rb`` - reading, binary mode
                * ``w`` - writing, text mode
                * ``wb`` - writing, binary mode

        Returns:
            A file-like interface, either :py:class:`BinaryIO` or
            :py:class:`TextIO` depending on whether the file was to be
            opened in binary or text mode. When writing, the return
            will be an instance of
            :py:class:`.item_io.DelfiniItemWriterBinary` or
            :py:class:`.item_io.DelfiniItemWriterText`.

        """
        if mode == "rb" or mode == "r":
            return item_io.read_item(
                mode,
                self.collection.id,
                self.collection.version_id,
                self.id,
                self._client,
            )

        elif mode == "wb":
            return item_io.DelfiniItemWriterBinary(
                self.collection.id, self.collection.version_id, self.id, self._client
            )

        elif mode == "w":
            return item_io.DelfiniItemWriterText(
                self.collection.id, self.collection.version_id, self.id, self._client
            )

    def table(self) -> pd.DataFrame:
        """Retrieve the tabular contents of this item as a :py:class:`pd.DataFrame`."""

        schema_metadata = {}

        def _get_data() -> Iterator[pa.RecordBatch]:
            nonlocal schema_metadata

            paginator = Paginator[TableData](
                collections_tables_get_table_data, self._client
            )
            kwargs = {
                "collection_id": self.collection.id,
                "version_id": self.collection.version_id,
                "table_name": self.id,
            }

            with tqdm(desc=f"Loading {self.path}") as pbar:
                for item_page in paginator.paginate(**kwargs):
                    schema_metadata = item_page.data_model.to_dict()
                    yield pa.RecordBatch.from_pylist(
                        [row.to_dict() for row in item_page.data],
                        schema=pyarrow_schema_from_json_schema(schema_metadata),
                    )
                    pbar.total = item_page.pagination.total_items
                    pbar.update(len(item_page.data))

        df = pa.Table.from_batches(_get_data()).to_pandas()

        props = schema_metadata["properties"]
        rename_map = {
            key: props[key]["description"]
            for key in props
            if "description" in props[key]
        }

        df.rename(columns=rename_map, inplace=True)

        return cast(pd.DataFrame, df)


class DelfiniFolder(DelfiniItem, FolderMixin):
    """Represents a folder within a collection.

    Note that folders cannot be opened as data streams, so the
    :py:meth:`open` method will always return :py:exc:`OSError`.

    """

    def __init__(
        self,
        collection: DelfiniCollection,
        in_folder: FolderMixin,
        model: Item,
        core: CoreAuthClient,
    ) -> None:
        DelfiniItem.__init__(self, collection, in_folder, model, core)
        FolderMixin.__init__(self, collection, in_folder._join_path(model.name), core)

    def __repr__(self) -> str:
        return (
            f"<DelfiniFolder: name={self.name} path={self.path} type={self.type}"
            f" id={self.id}>"
        )

    @overload
    def open(self, mode: item_io.ItemBinaryModes) -> BinaryIO:
        ...

    @overload
    def open(self, mode: item_io.ItemTextModes) -> TextIO:
        ...

    def open(self, mode: item_io.ItemIOModes = "r") -> Union[BinaryIO, TextIO]:
        """N/A - folders cannot be opened.

        Raises: :py:exc:`OSError` always
        """
        raise OSError("cannot open folders")



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.event_metric_dimensions import EventMetricDimensions
from ..models.event_metric_type import EventMetricType
from ..models.metric_agg_func import MetricAggFunc
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="EventMetric")


@_attrs_define
class EventMetric:
    """EventMetric model

    Attributes:
        agg_func (MetricAggFunc):
        description (str):
        event_name (str):
        name (str):
        dimensions (Union[Unset, EventMetricDimensions]):
        num_partitions (Union[Unset, int]):
        partition_by (Union[Unset, str]):
        type (Union[Unset, EventMetricType]):  Default: EventMetricType.EVENT.
    """

    agg_func: MetricAggFunc
    description: str
    event_name: str
    name: str
    dimensions: Union[Unset, "EventMetricDimensions"] = UNSET
    num_partitions: Union[Unset, int] = UNSET
    partition_by: Union[Unset, str] = UNSET
    type: Union[Unset, EventMetricType] = EventMetricType.EVENT

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        agg_func = self.agg_func.value
        description = self.description
        event_name = self.event_name
        name = self.name
        dimensions: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.dimensions, Unset):
            dimensions = self.dimensions.to_dict()
        num_partitions = self.num_partitions
        partition_by = self.partition_by
        type: Union[Unset, str] = UNSET
        if not isinstance(self.type, Unset):
            type = self.type.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "agg_func": agg_func,
                "description": description,
                "event_name": event_name,
                "name": name,
            }
        )
        if dimensions is not UNSET:
            field_dict["dimensions"] = dimensions
        if num_partitions is not UNSET:
            field_dict["num_partitions"] = num_partitions
        if partition_by is not UNSET:
            field_dict["partition_by"] = partition_by
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`EventMetric` from a dict"""
        d = src_dict.copy()
        agg_func = MetricAggFunc(d.pop("agg_func"))

        description = d.pop("description")

        event_name = d.pop("event_name")

        name = d.pop("name")

        _dimensions = d.pop("dimensions", UNSET)
        dimensions: Union[Unset, EventMetricDimensions]
        if isinstance(_dimensions, Unset):
            dimensions = UNSET
        else:
            dimensions = EventMetricDimensions.from_dict(_dimensions)

        num_partitions = d.pop("num_partitions", UNSET)

        partition_by = d.pop("partition_by", UNSET)

        _type = d.pop("type", UNSET)
        type: Union[Unset, EventMetricType]
        if isinstance(_type, Unset):
            type = UNSET
        else:
            type = EventMetricType(_type)

        event_metric = cls(
            agg_func=agg_func,
            description=description,
            event_name=event_name,
            name=name,
            dimensions=dimensions,
            num_partitions=num_partitions,
            partition_by=partition_by,
            type=type,
        )

        return event_metric



from .accounts import AccountsCommands
from .admin import AdminCommands
from .auth import AuthCommands
from .base_commands import BaseCommands
from .cde import CdeCommands
from .collection import CollectionCommands
from .groups import GroupsCommands
from .pdd import PddCommands
from .tasks import TasksCommands
from .users import UsersCommands


class Commands(BaseCommands):
    accounts = AccountsCommands
    admin = AdminCommands
    auth = AuthCommands
    cdeset = CdeCommands
    collection = CollectionCommands
    groups = GroupsCommands
    pdd = PddCommands
    tasks = TasksCommands
    users = UsersCommands



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define


T = TypeVar("T", bound="UpdateAccountPages")


@_attrs_define
class UpdateAccountPages:
    """UpdateAccountPages model

    Attributes:
        collection_id (str):
        item_ids (List[str]):
        version_id (str):
    """

    collection_id: str
    item_ids: List[str]
    version_id: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        collection_id = self.collection_id
        item_ids = self.item_ids

        version_id = self.version_id

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "collection_id": collection_id,
                "item_ids": item_ids,
                "version_id": version_id,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`UpdateAccountPages` from a dict"""
        d = src_dict.copy()
        collection_id = d.pop("collection_id")

        item_ids = cast(List[str], d.pop("item_ids"))

        version_id = d.pop("version_id")

        update_account_pages = cls(
            collection_id=collection_id,
            item_ids=item_ids,
            version_id=version_id,
        )

        return update_account_pages



from pydelfini.delfini_core.api.auth import auth_get_session
from pydelfini.delfini_core.api.user import user_update_user
from pydelfini.delfini_core.models import AuthGetSessionResponse200
from pydelfini.delfini_core.models import UpdateUser
from pydelfini.delfini_core.models import UpdateUserMetadata

from .base_commands import BaseCommands


class AuthCommands(BaseCommands):
    """Authentication and authorization operations"""

    def whoami(self) -> None:
        """Information about your current session"""
        session = auth_get_session.sync(client=self.core)
        if not isinstance(session, AuthGetSessionResponse200):
            raise Exception("could not read session")

        self._output(session.to_dict())

    @BaseCommands._with_arg("key")
    @BaseCommands._with_arg("value")
    def set_metadata(self) -> None:
        session = auth_get_session.sync(client=self.core)
        if not isinstance(session, AuthGetSessionResponse200):
            raise Exception("could not read session")

        user_name = session.user.identity.user_name
        metadata: dict[str, str] = {}

        # TODO: fetch existing metadata so that this is not destructive
        # response = user_get_user.sync(user_name)
        # ...
        metadata[self.args.key] = self.args.value

        user_update_user.sync(
            user_name,
            body=UpdateUser(metadata=UpdateUserMetadata.from_dict(metadata)),
            client=self.core,
        )



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.new_project_metadata import NewProjectMetadata
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="NewProject")


@_attrs_define
class NewProject:
    """NewProject model

    Attributes:
        account_id (str):
        description (str):
        name (str):
        metadata (Union[Unset, NewProjectMetadata]):
    """

    account_id: str
    description: str
    name: str
    metadata: Union[Unset, "NewProjectMetadata"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        account_id = self.account_id
        description = self.description
        name = self.name
        metadata: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "account_id": account_id,
                "description": description,
                "name": name,
            }
        )
        if metadata is not UNSET:
            field_dict["metadata"] = metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`NewProject` from a dict"""
        d = src_dict.copy()
        account_id = d.pop("account_id")

        description = d.pop("description")

        name = d.pop("name")

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, NewProjectMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = NewProjectMetadata.from_dict(_metadata)

        new_project = cls(
            account_id=account_id,
            description=description,
            name=name,
            metadata=metadata,
        )

        return new_project



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.metric_log_entry import MetricLogEntry


T = TypeVar("T", bound="MetricLog")


@_attrs_define
class MetricLog:
    """MetricLog model

    Attributes:
        events (List['MetricLogEntry']):
    """

    events: List["MetricLogEntry"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        events = []
        for events_item_data in self.events:
            events_item = events_item_data.to_dict()
            events.append(events_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "events": events,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`MetricLog` from a dict"""
        d = src_dict.copy()
        events = []
        _events = d.pop("events")
        for events_item_data in _events:
            events_item = MetricLogEntry.from_dict(events_item_data)

            events.append(events_item)

        metric_log = cls(
            events=events,
        )

        return metric_log



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.datastore_type import DatastoreType


T = TypeVar("T", bound="Datastore")


@_attrs_define
class Datastore:
    """Public view on a datastore

    Attributes:
        enabled (bool):
        name (str):
        num_objects (int):
        size_bytes (int):
        type (DatastoreType):
        url (str):
    """

    enabled: bool
    name: str
    num_objects: int
    size_bytes: int
    type: DatastoreType
    url: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        enabled = self.enabled
        name = self.name
        num_objects = self.num_objects
        size_bytes = self.size_bytes
        type = self.type.value
        url = self.url

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "enabled": enabled,
                "name": name,
                "num_objects": num_objects,
                "size_bytes": size_bytes,
                "type": type,
                "url": url,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`Datastore` from a dict"""
        d = src_dict.copy()
        enabled = d.pop("enabled")

        name = d.pop("name")

        num_objects = d.pop("num_objects")

        size_bytes = d.pop("size_bytes")

        type = DatastoreType(d.pop("type"))

        url = d.pop("url")

        datastore = cls(
            enabled=enabled,
            name=name,
            num_objects=num_objects,
            size_bytes=size_bytes,
            type=type,
            url=url,
        )

        return datastore



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.group_member import GroupMember
from ..models.group_members_users import GroupMembersUsers
from ..models.pagination import Pagination
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="GroupMembers")


@_attrs_define
class GroupMembers:
    """GroupMembers model

    Attributes:
        members (List['GroupMember']):
        pagination (Pagination):
        users (Union[Unset, GroupMembersUsers]):
    """

    members: List["GroupMember"]
    pagination: "Pagination"
    users: Union[Unset, "GroupMembersUsers"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        members = []
        for members_item_data in self.members:
            members_item = members_item_data.to_dict()
            members.append(members_item)

        pagination = self.pagination.to_dict()
        users: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.users, Unset):
            users = self.users.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "members": members,
                "pagination": pagination,
            }
        )
        if users is not UNSET:
            field_dict["users"] = users

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`GroupMembers` from a dict"""
        d = src_dict.copy()
        members = []
        _members = d.pop("members")
        for members_item_data in _members:
            members_item = GroupMember.from_dict(members_item_data)

            members.append(members_item)

        pagination = Pagination.from_dict(d.pop("pagination"))

        _users = d.pop("users", UNSET)
        users: Union[Unset, GroupMembersUsers]
        if isinstance(_users, Unset):
            users = UNSET
        else:
            users = GroupMembersUsers.from_dict(_users)

        group_members = cls(
            members=members,
            pagination=pagination,
            users=users,
        )

        return group_members



from pydelfini.delfini_core import Paginator
from pydelfini.delfini_core.api.account import account_create_account
from pydelfini.delfini_core.api.account import account_delete_account
from pydelfini.delfini_core.api.account import account_list_accounts
from pydelfini.delfini_core.api.account import account_update_account_pages
from pydelfini.delfini_core.api.items import collections_items_list_items
from pydelfini.delfini_core.models import AccountCreateAccountBody
from pydelfini.delfini_core.models import AccountCreateAccountBodyMetadata
from pydelfini.delfini_core.models import CollectionsItemsListItemsResponse200
from pydelfini.delfini_core.models import ItemType
from pydelfini.delfini_core.models import UpdateAccountPages

from .base_commands import BaseCommands


class AccountsCommands(BaseCommands):
    """Operations on accounts"""

    @BaseCommands._with_arg("name")
    @BaseCommands._with_arg("--description", default="")
    def new(self) -> None:
        """Create a new account."""
        rv = account_create_account.sync(
            body=AccountCreateAccountBody(
                name=self.args.name,
                metadata=AccountCreateAccountBodyMetadata.from_dict(
                    {"x-description": self.args.description}
                ),
            ),
            client=self.core,
        )

        self._output(rv.to_dict())

    def list(self) -> None:
        """List all accounts."""
        rv = account_list_accounts.sync(client=self.core)

        self._output(rv.to_dict())

    @BaseCommands._with_arg("account_id")
    def delete(self) -> None:
        """Remove an account."""
        rv = account_delete_account.sync(
            account_id=self.args.account_id, client=self.core
        )
        assert rv

    @BaseCommands._with_arg("account_id")
    @BaseCommands._with_arg("collection_id")
    @BaseCommands._with_arg("--version-id", default="LIVE")
    def update_space(self) -> None:
        """Copy pages from the designated collection into the account space."""
        # first, we need a list of item IDs
        pager = Paginator[CollectionsItemsListItemsResponse200](
            collections_items_list_items, self.core
        )
        item_ids: list[str] = []
        for page in pager.paginate(
            collection_id=self.args.collection_id,
            version_id=self.args.version_id,
            type=[ItemType.FILE],
        ):
            item_ids.extend(i.id for i in page.items)

        account_update_account_pages.sync(
            account_id=self.args.account_id,
            body=UpdateAccountPages(
                collection_id=self.args.collection_id,
                version_id=self.args.version_id,
                item_ids=item_ids,
            ),
            client=self.core,
        )



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.data_dictionary import DataDictionary
from ..models.pagination import Pagination


T = TypeVar("T", bound="CollectionsDictionariesListDictionariesResponse200")


@_attrs_define
class CollectionsDictionariesListDictionariesResponse200:
    """CollectionsDictionariesListDictionariesResponse200 model

    Attributes:
        dictionaries (List['DataDictionary']):
        pagination (Pagination):
    """

    dictionaries: List["DataDictionary"]
    pagination: "Pagination"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        dictionaries = []
        for dictionaries_item_data in self.dictionaries:
            dictionaries_item = dictionaries_item_data.to_dict()
            dictionaries.append(dictionaries_item)

        pagination = self.pagination.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "dictionaries": dictionaries,
                "pagination": pagination,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`CollectionsDictionariesListDictionariesResponse200` from a dict"""
        d = src_dict.copy()
        dictionaries = []
        _dictionaries = d.pop("dictionaries")
        for dictionaries_item_data in _dictionaries:
            dictionaries_item = DataDictionary.from_dict(dictionaries_item_data)

            dictionaries.append(dictionaries_item)

        pagination = Pagination.from_dict(d.pop("pagination"))

        collections_dictionaries_list_dictionaries_response_200 = cls(
            dictionaries=dictionaries,
            pagination=pagination,
        )

        return collections_dictionaries_list_dictionaries_response_200



from enum import Enum


class MetadataGetFieldGroupsByResourceResourceType(str, Enum):
    ACCOUNT = "account"
    COLLECTION = "collection"
    COLLECTION_ITEM = "collection-item"
    GROUP = "group"
    USER = "user"

    def __str__(self) -> str:
        return str(self.value)



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field

from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="SystemConfigurationSessions")


@_attrs_define
class SystemConfigurationSessions:
    """SystemConfigurationSessions model

    Attributes:
        persistent (Union[Unset, bool]):  Default: True.
        persistent_lifetime (Union[Unset, int]):  Default: 1209600.
    """

    persistent: Union[Unset, bool] = True
    persistent_lifetime: Union[Unset, int] = 1209600
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        persistent = self.persistent
        persistent_lifetime = self.persistent_lifetime

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if persistent is not UNSET:
            field_dict["persistent"] = persistent
        if persistent_lifetime is not UNSET:
            field_dict["persistent_lifetime"] = persistent_lifetime

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SystemConfigurationSessions` from a dict"""
        d = src_dict.copy()
        persistent = d.pop("persistent", UNSET)

        persistent_lifetime = d.pop("persistent_lifetime", UNSET)

        system_configuration_sessions = cls(
            persistent=persistent,
            persistent_lifetime=persistent_lifetime,
        )

        system_configuration_sessions.additional_properties = d
        return system_configuration_sessions

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.collections_tables_get_table_data_elements_response_200_element_map import (
    CollectionsTablesGetTableDataElementsResponse200ElementMap,
)
from ..models.collections_tables_get_table_data_elements_response_200_error_map import (
    CollectionsTablesGetTableDataElementsResponse200ErrorMap,
)


T = TypeVar("T", bound="CollectionsTablesGetTableDataElementsResponse200")


@_attrs_define
class CollectionsTablesGetTableDataElementsResponse200:
    """CollectionsTablesGetTableDataElementsResponse200 model

    Attributes:
        element_map (CollectionsTablesGetTableDataElementsResponse200ElementMap):
        error_map (CollectionsTablesGetTableDataElementsResponse200ErrorMap):
    """

    element_map: "CollectionsTablesGetTableDataElementsResponse200ElementMap"
    error_map: "CollectionsTablesGetTableDataElementsResponse200ErrorMap"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        element_map = self.element_map.to_dict()
        error_map = self.error_map.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "elementMap": element_map,
                "errorMap": error_map,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`CollectionsTablesGetTableDataElementsResponse200` from a dict"""
        d = src_dict.copy()
        element_map = (
            CollectionsTablesGetTableDataElementsResponse200ElementMap.from_dict(
                d.pop("elementMap")
            )
        )

        error_map = CollectionsTablesGetTableDataElementsResponse200ErrorMap.from_dict(
            d.pop("errorMap")
        )

        collections_tables_get_table_data_elements_response_200 = cls(
            element_map=element_map,
            error_map=error_map,
        )

        return collections_tables_get_table_data_elements_response_200



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.metadata_field_group_resource import MetadataFieldGroupResource
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="MetadataFieldGroup")


@_attrs_define
class MetadataFieldGroup:
    """MetadataFieldGroup model

    Attributes:
        field_names (List[str]):
        id (str): Unique ID for the group
        resource (MetadataFieldGroupResource):
        title (str):
        description (Union[Unset, str]):
        edit_page (Union[Unset, bool]):
        view_position (Union[Unset, str]):
        visibility (Union[Unset, str]):
    """

    field_names: List[str]
    id: str
    resource: MetadataFieldGroupResource
    title: str
    description: Union[Unset, str] = UNSET
    edit_page: Union[Unset, bool] = UNSET
    view_position: Union[Unset, str] = UNSET
    visibility: Union[Unset, str] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        field_names = self.field_names

        id = self.id
        resource = self.resource.value
        title = self.title
        description = self.description
        edit_page = self.edit_page
        view_position = self.view_position
        visibility = self.visibility

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "field_names": field_names,
                "id": id,
                "resource": resource,
                "title": title,
            }
        )
        if description is not UNSET:
            field_dict["description"] = description
        if edit_page is not UNSET:
            field_dict["edit_page"] = edit_page
        if view_position is not UNSET:
            field_dict["view_position"] = view_position
        if visibility is not UNSET:
            field_dict["visibility"] = visibility

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`MetadataFieldGroup` from a dict"""
        d = src_dict.copy()
        field_names = cast(List[str], d.pop("field_names"))

        id = d.pop("id")

        resource = MetadataFieldGroupResource(d.pop("resource"))

        title = d.pop("title")

        description = d.pop("description", UNSET)

        edit_page = d.pop("edit_page", UNSET)

        view_position = d.pop("view_position", UNSET)

        visibility = d.pop("visibility", UNSET)

        metadata_field_group = cls(
            field_names=field_names,
            id=id,
            resource=resource,
            title=title,
            description=description,
            edit_page=edit_page,
            view_position=view_position,
            visibility=visibility,
        )

        return metadata_field_group



from pydelfini.delfini_core.api.user import user_delete_user
from pydelfini.delfini_core.api.user import user_get_users
from pydelfini.delfini_core.api.user import user_update_user_admin
from pydelfini.delfini_core.models import UserAdminUpdate
from pydelfini.delfini_core.models import UserGetUsersResponse200
from pydelfini.delfini_core.paginator import Paginator

from .base_commands import BaseCommands


class UsersCommands(BaseCommands):
    """Operations on users"""

    def list(self) -> None:
        """List users."""
        pager = Paginator[UserGetUsersResponse200](user_get_users, self.core)

        data = []
        for page in pager.paginate():
            data.extend([u.to_dict() for u in page.users])

        self._output(data, preferred_type="table")

    @BaseCommands._with_arg("username")
    def delete(self) -> None:
        """Delete a user."""
        user_delete_user.sync(self.args.username, client=self.core)

    @BaseCommands._with_arg("username")
    def disable(self) -> None:
        """Disable a user."""
        resp = user_update_user_admin.sync(
            self.args.username,
            body=UserAdminUpdate(is_disabled=True),
            client=self.core,
        )
        self._output(resp.to_dict())

    @BaseCommands._with_arg("username")
    def enable(self) -> None:
        """Enable a user."""
        resp = user_update_user_admin.sync(
            self.args.username,
            body=UserAdminUpdate(is_disabled=False),
            client=self.core,
        )
        self._output(resp.to_dict())



from pydelfini.delfini_core.api.admin import admin_grant_admin
from pydelfini.delfini_core.api.admin import admin_list_admins
from pydelfini.delfini_core.models import AdminGrantAdminAdminListRequest
from pydelfini.delfini_core.models import AdminGrants
from pydelfini.delfini_core.models import Operations

from .base_commands import BaseCommands
from .config import ConfigCommands


class AdminCommands(BaseCommands):
    """Administrative operations"""

    config = ConfigCommands

    def list_admins(self) -> None:
        """List system admins and privileges."""
        rv = admin_list_admins.sync(client=self.core)

        self._output(
            [
                {"subject_id": r.subject_id, "operations": ",".join(r.operations)}
                for r in rv.grants
            ],
            preferred_type="table",
        )

        if rv.operations:
            print()
            print("Operations:")
            for op in rv.operations:
                print("-", op)

    @BaseCommands._with_arg("subject_id")
    @BaseCommands._with_arg("-o", "--operation")
    def grant_admin(self) -> None:
        """Grant admin privilege to a subject ID."""
        admins = admin_list_admins.sync(client=self.core)

        if self.args.operation:
            new_operations = {Operations(self.args.operation)}
        else:
            new_operations = set(admins.operations or [])

        for grant in admins.grants:
            if grant.subject_id == self.args.subject_id:
                grant.operations = list(set(grant.operations) | new_operations)

                break
        else:
            admins.grants.append(
                AdminGrants(
                    subject_id=self.args.subject_id, operations=list(new_operations)
                )
            )

        rv = admin_grant_admin.sync(
            body=AdminGrantAdminAdminListRequest(grants=admins.grants), client=self.core
        )
        assert rv

        print("Updated", self.args.subject_id, " -- now admins are:")
        self.list_admins()

    @BaseCommands._with_arg("subject_id")
    def revoke_admin(self) -> None:
        """Revoke all admin privileges from a user."""
        admins = admin_list_admins.sync(client=self.core)

        new_grants = [g for g in admins.grants if g.subject_id != self.args.subject_id]
        if new_grants == admins.grants:
            raise ValueError(f"{self.args.subject_id} is not listed in admins")

        admin_grant_admin.sync(
            body=AdminGrantAdminAdminListRequest(grants=new_grants), client=self.core
        )



from enum import Enum


class MetricsQueryEventDataPeriod(str, Enum):
    VALUE_0 = "5-minutes"
    VALUE_1 = "15-minutes"
    VALUE_2 = "1-hour"
    VALUE_3 = "1-day"
    VALUE_4 = "1-week"
    VALUE_5 = "30-days"

    def __str__(self) -> str:
        return str(self.value)



import datetime
from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define
from dateutil.parser import isoparse

from ..models.task_schedule_def import TaskScheduleDef


T = TypeVar("T", bound="TasksSchedulesSchedulesItem")


@_attrs_define
class TasksSchedulesSchedulesItem:
    """TasksSchedulesSchedulesItem model

    Attributes:
        definition (TaskScheduleDef):
        next_run_time (datetime.datetime):
    """

    definition: "TaskScheduleDef"
    next_run_time: datetime.datetime

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        definition = self.definition.to_dict()
        next_run_time = self.next_run_time.isoformat()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "definition": definition,
                "next_run_time": next_run_time,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`TasksSchedulesSchedulesItem` from a dict"""
        d = src_dict.copy()
        definition = TaskScheduleDef.from_dict(d.pop("definition"))

        next_run_time = isoparse(d.pop("next_run_time"))

        tasks_schedules_schedules_item = cls(
            definition=definition,
            next_run_time=next_run_time,
        )

        return tasks_schedules_schedules_item



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.data_element_permissible_values_date_time_format_date_time_format import (
    DataElementPermissibleValuesDateTimeFormatDateTimeFormat,
)


T = TypeVar("T", bound="DataElementPermissibleValuesDateTimeFormat")


@_attrs_define
class DataElementPermissibleValuesDateTimeFormat:
    """DataElementPermissibleValuesDateTimeFormat model

    Attributes:
        date_time_format (DataElementPermissibleValuesDateTimeFormatDateTimeFormat): Controls on the date, time or
            duration domain of the data.
    """

    date_time_format: "DataElementPermissibleValuesDateTimeFormatDateTimeFormat"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        date_time_format = self.date_time_format.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "dateTimeFormat": date_time_format,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`DataElementPermissibleValuesDateTimeFormat` from a dict"""
        d = src_dict.copy()
        date_time_format = (
            DataElementPermissibleValuesDateTimeFormatDateTimeFormat.from_dict(
                d.pop("dateTimeFormat")
            )
        )

        data_element_permissible_values_date_time_format = cls(
            date_time_format=date_time_format,
        )

        return data_element_permissible_values_date_time_format



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.multipart_upload_details_target_urls import (
    MultipartUploadDetailsTargetUrls,
)


T = TypeVar("T", bound="MultipartUploadDetails")


@_attrs_define
class MultipartUploadDetails:
    """MultipartUploadDetails model

    Attributes:
        target_urls (MultipartUploadDetailsTargetUrls): A mapping of byte ranges ("<start>-<end>", inclusive) to target
            URLs.
        upload_id (str):
        upload_key (str):
        url (str):
    """

    target_urls: "MultipartUploadDetailsTargetUrls"
    upload_id: str
    upload_key: str
    url: str

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        target_urls = self.target_urls.to_dict()
        upload_id = self.upload_id
        upload_key = self.upload_key
        url = self.url

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "targetUrls": target_urls,
                "uploadId": upload_id,
                "uploadKey": upload_key,
                "url": url,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`MultipartUploadDetails` from a dict"""
        d = src_dict.copy()
        target_urls = MultipartUploadDetailsTargetUrls.from_dict(d.pop("targetUrls"))

        upload_id = d.pop("uploadId")

        upload_key = d.pop("uploadKey")

        url = d.pop("url")

        multipart_upload_details = cls(
            target_urls=target_urls,
            upload_id=upload_id,
            upload_key=upload_key,
            url=url,
        )

        return multipart_upload_details



from datetime import datetime

import pytest
from pydelfini.delfini_core import paginator
from pydelfini.delfini_core.api.collections import collections_get_collections


def test_paginate(httpx_mock, client):
    def make_collection(i):
        return {
            "id": i,
            "versionId": "LIVE",
            "name": i,
            "description": "",
            "metadata": {},
            "accessLevel": "private",
            "account_id": "aaa",
            "createdOn": datetime.now().isoformat(),
        }

    httpx_mock.add_response(
        url="https://delfini.test/api/v1/collections?page_size=50",
        json={
            "collections": [
                make_collection("a"),
                make_collection("b"),
                make_collection("c"),
                make_collection("d"),
                make_collection("e"),
            ],
            "pagination": {
                "items_per_page": 5,
                "total_items": 7,
                "next_page_url": "https://delfini.test/api/v1/collections?token=xyz",
            },
        },
    )
    httpx_mock.add_response(
        url="https://delfini.test/api/v1/collections?token=xyz",
        json={
            "collections": [
                make_collection("f"),
                make_collection("g"),
            ],
            "pagination": {
                "items_per_page": 5,
                "total_items": 7,
                "next_page_url": None,
            },
        },
    )

    pager = paginator.Paginator(collections_get_collections, client)
    p = pager.paginate()

    page_one = next(p)
    assert len(page_one.collections) == 5

    page_two = next(p)
    assert len(page_two.collections) == 2

    with pytest.raises(StopIteration):
        next(p)



import argparse
import ast
import os
import random
import re
import shlex
from pathlib import Path
from urllib.request import urlopen

import pytest
from pydelfini.client import make_base_url
from pydelfini.dcmd import cli
from pydelfini.dcmd.commands import Commands as PcmdCommands
from pydelfini.delfini_core import Client as CoreClient
from pydelfini.delfini_core.api.user import user_create_user
from pydelfini.delfini_core.login import Login
from pydelfini.delfini_core.login import LoginError
from pydelfini.delfini_core.login import to_token_file
from pydelfini.delfini_core.models import NewUser
from yaml import dump as yaml_dump
from yaml import safe_load


# The tests in this test suite are designed to be run against a local
# development instance of Delfini hosted at localhost:8170, similar to
# the Cypress tests in `delfini-nuxt`. In order to be able to run these
# tests, be sure to have the local development instance running in the
# background with the command:
#
#    flask run --debug
#


@pytest.fixture
def local_user(tmp_path):
    # try to log in with the credentials
    username, password = "dcmd.tester", "insecure"
    base_url = make_base_url("localhost:8170", "http")
    try:
        authenticated = Login(CoreClient(base_url=base_url)).with_password(
            username, password
        )
    except LoginError:
        print("Login failed, trying to create new user...")
        user_create_user.sync(
            body=NewUser(
                given_name="dcmd",
                last_name="tester",
                user_email="john.doe@example.com",
                user_name=username,
                password=password,
            ),
            client=CoreClient(base_url=base_url),
        )
        authenticated = Login(CoreClient(base_url=base_url)).with_password(
            username, password
        )

    token_file = tmp_path / "token"
    to_token_file(authenticated, token_file)

    yield str(token_file)


def random_id():
    return f"rr{random.randint(0, 1048576):05x}rr"


def mask(rv):
    # mask UUIDs
    rv = re.sub(
        r"[0-9a-f]{8}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{4}-[0-9a-f]{12}",
        "00000000-0000-0000-0000-000000000000",
        rv,
    )
    # mask timestamps
    rv = re.sub(
        r"\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}(\.[0-9+:]*)?",
        "1970-01-02T03:04:05.000000",
        rv,
    )
    # mask task worker
    rv = re.sub(r"worker: .*", "worker: dcmd-test-mask", rv)
    # mask random ID
    rv = re.sub(r"rr[0-9a-f]{5}rr", "random-id", rv)
    # mask bundle ID
    rv = re.sub(r"b-[a-zA-Z0-9]{8}", "b-00000000", rv)

    return rv


@pytest.fixture
def dcmd(local_user, capsys):
    def _cmd(args, masked=True):
        if isinstance(args, str):
            args = shlex.split(args)
        cli.main(["-H", "localhost:8170", "--insecure", "-k", local_user] + args)

        outv = capsys.readouterr()
        if not outv.err:
            rv = outv.out
            if masked:
                rv = mask(rv)
            return rv

        return outv

    yield _cmd


@pytest.fixture
def test_target_user(dcmd):
    response = user_create_user.sync(
        body=NewUser(
            given_name="dcmd",
            last_name="test target",
            user_email="john.doe@example.com",
            user_name="dcmd.test.target",
            password=random_id(),
        ),
        client=CoreClient(base_url=make_base_url("localhost:8170", "http")),
    )

    yield response.subject_id

    dcmd("users delete dcmd.test.target")


@pytest.fixture
def scratch_collection(dcmd):
    collection_name = f"dcmd-test.{random_id()}"
    rv = safe_load(dcmd(f"collection new {collection_name}", masked=False))

    yield collection_name, rv["id"]

    dcmd(f"collection delete {rv['id']}")


def test_dcmd_auth_whoami(dcmd, snapshot):
    dcmd('auth set-metadata x-tos \'{"toc":3,"coc":6,"priv":3,"com":1}\'')

    assert dcmd("auth whoami") == snapshot


# NOTE: If the following tests are failing with "Unexpected status
# code: 403", you should grant the `dcmd.tester` user full admin
# access to your local development instance with the command:
#
#   flask admin grant dcmd.tester
#


def test_dcmd_tasks(dcmd, snapshot):
    stats = safe_load(dcmd("tasks stats"))
    assert stats.keys() == {"queued", "results", "schedules"}

    schedules = safe_load(dcmd("tasks schedules"))["schedules"]
    assert len(schedules) >= 2

    results = safe_load(dcmd("tasks results --idemkey tasks-backends-sqlite-vacuum"))[
        "results"
    ]
    if len(results) > 0:
        assert results[0] == snapshot


def test_dcmd_users(dcmd, test_target_user):
    rv = dcmd("users list")
    assert set(rv.splitlines()[0].split()) == {
        "account_id",
        "created_on",
        "metadata",
        "subject_id",
        "user_name",
        "visibility_level",
        "given_name",
        "groups",
        "last_name",
        "user_email",
    }
    assert "dcmd.test.target" in rv

    # disable and enable
    rv = safe_load(dcmd("users disable dcmd.test.target"))
    assert rv["disabled_on"] is not None

    rv = safe_load(dcmd("users enable dcmd.test.target"))
    assert rv["disabled_on"] is None


def test_dcmd_groups(dcmd, test_target_user, snapshot):
    group_name = f"test-dcmd-group.{random_id()}"
    new_group = dcmd(f"groups new {group_name}", masked=False)
    group_id = safe_load(new_group)["id"]
    assert mask(new_group) == snapshot

    groups = safe_load(dcmd("groups list"))["groups"]
    assert safe_load(mask(new_group)) in groups

    dcmd(f"groups join {group_id} {test_target_user}")

    assert len(safe_load(dcmd(f"groups list-members {group_id}"))["members"]) == 2

    dcmd(f"groups update --name new-test-group {group_id}")

    dcmd(f"groups delete {group_id}")


def test_dcmd_accounts(dcmd, snapshot, tmp_path, scratch_collection):
    collection_name, collection_id = scratch_collection

    account_name = f"test-dcmd-account.{random_id()}"
    new_account = dcmd(f"accounts new {account_name} --description heyyo", masked=False)
    account_id = safe_load(new_account)["id"]
    assert mask(new_account) == snapshot

    accounts = safe_load(dcmd("accounts list", masked=False))["accounts"]
    assert safe_load(new_account) in accounts

    # update space
    index_file = tmp_path / "index.md"
    with open(index_file, "w") as fp:
        fp.write("# Hello world!\n")
    dcmd(f"collection upload {collection_name} {index_file}")
    dcmd(f"accounts update-space {account_id} {collection_id}")
    response = urlopen(
        f"http://localhost:8170/api/v1/account/{account_id}/pages/index.md"
    )
    assert response.code == 200

    dcmd(f"accounts delete {account_id}")

    accounts = safe_load(dcmd("accounts list", masked=False))["accounts"]
    assert safe_load(new_account) not in accounts


def test_dcmd_admin_grant(dcmd, test_target_user):
    dcmd(f"admin grant-admin -o group {test_target_user}")

    rv = dcmd("admin list-admins", masked=False)
    assert test_target_user in rv

    dcmd(f"admin revoke-admin {test_target_user}")

    rv = dcmd("admin list-admins", masked=False)
    assert test_target_user not in rv


def test_dcmd_admin_config(dcmd, tmp_path, capsys):
    rv = safe_load(dcmd("admin config get"))
    assert "datastores" in rv
    assert "authentication" in rv

    dcmd(
        "admin config set datastores.dcmd_test"
        + f' \'{{"enabled": false, "path": "{tmp_path}", "type": "local"}}\''
    )

    rv = safe_load(dcmd("admin config get", masked=False))
    assert "dcmd_test" in rv["datastores"]

    newdir = tmp_path / "new"
    rv["datastores"]["dcmd_test"]["path"] = str(newdir)

    config_fn = tmp_path / "config.yaml"
    with open(config_fn, "w") as fp:
        yaml_dump(rv, fp)

    dcmd(f"admin config put {config_fn}")

    rv = safe_load(dcmd("admin config get"))
    assert "dcmd_test" in rv["datastores"]
    assert rv["datastores"]["dcmd_test"]["path"] == str(newdir)

    dcmd("admin config delete datastores.dcmd_test")

    rv = safe_load(dcmd("admin config get"))
    assert "dcmd_test" not in rv["datastores"]


def test_dcmd_collection(dcmd, tmp_path, scratch_collection):
    collection_name, collection_id = scratch_collection
    rv = dcmd("collection list", masked=False)
    assert collection_name in rv
    assert collection_id in rv

    test_file = tmp_path / "test.txt"
    with open(test_file, "w") as fp:
        fp.write("hello world this is a test\n")

    dcmd(f"collection upload {collection_name} {test_file}")

    test_file.unlink()

    old_cwd = os.getcwd()
    os.chdir(tmp_path)
    dcmd(f"collection download {collection_name} test.txt")
    os.chdir(old_cwd)

    assert test_file.exists()


def test_dcmd_pdd(dcmd, scratch_collection, snapshot):
    collection_name, collection_id = scratch_collection
    common_pdd_fn = Path(__file__).parent / "common_pdd.json"

    rv = dcmd(f"pdd upload {collection_id} {common_pdd_fn}", masked=False)
    assert mask(rv) == snapshot
    safe_load(rv)["id"]

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot

    dcmd(f"pdd set-order {collection_name} common_pdd.json 6fr70ICV O8zPEsiM")

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot

    assert (
        dcmd(f"pdd bundle {collection_name} common_pdd.json MyBundle O8zPEsiM 6fr70ICV")
        == snapshot
    )

    assert dcmd(f"pdd list {collection_name} common_pdd.json") == snapshot


def test_dcmd_cdeset(dcmd, scratch_collection):
    collection_name, collection_id = scratch_collection
    common_pdd_fn = Path(__file__).parent / "common_pdd.json"

    rv = dcmd(f"pdd upload {collection_id} {common_pdd_fn}", masked=False)
    item_id = safe_load(rv)["id"]

    cdeset_name = f"dcmd.{random_id()}"
    dcmd(f"cdeset new {cdeset_name} 'dcmd test cde set'")

    rv = dcmd("cdeset list", masked=False)
    assert cdeset_name in rv

    dcmd(f"cdeset copy-from-pdd {cdeset_name} {collection_id} LIVE {item_id}")

    rv = safe_load(dcmd("cdeset list", masked=False))
    cdeset = {"f": c for c in rv["cdesets"] if c["name"] == cdeset_name}.get("f")
    assert cdeset["num_cdes"] == 2
    assert cdeset["description"] == "dcmd test cde set"

    dcmd(f"cdeset delete {cdeset_name}")


###
### Coverage Verification
###


class CoverageScanner(ast.NodeVisitor):
    def __init__(self):
        super().__init__()
        self.actions = {}

    def register_actions(self, parser, actlevel=None):
        if actlevel is None:
            actlevel = self.actions

        actions = parser._get_positional_actions()
        if actions and actions[0].choices:
            for arg, subparser in actions[0].choices.items():
                self.register_actions(subparser, actlevel.setdefault(arg, {}))

    def coverage_needed(self):
        needed = set()

        def _walk(d, s=None):
            if s is None:
                s = []
            if "SEEN" in d:
                return
            if not d:
                needed.add(" ".join(s))
            for k in d:
                _walk(d[k], s + [k])

        _walk(self.actions)
        return needed

    def visit_Call(self, node):
        if not isinstance(node.func, ast.Name):
            return self.generic_visit(node)
        if node.func.id != "dcmd":
            return self.generic_visit(node)
        if not node.args:
            return self.generic_visit(node)

        def _resolve(arg):
            if isinstance(arg, ast.Constant):
                return arg.value
            elif isinstance(arg, ast.JoinedStr):
                return _resolve(arg.values[0])
            elif isinstance(arg, ast.BinOp):
                return _resolve(arg.left)
            else:
                print(ast.dump(node, indent=2))
                raise Exception(
                    f"unexpected dcmd call on line {node.lineno}: {ast.unparse(node)}"
                )

        dcmd_arg = _resolve(node.args[0])

        actlevel = self.actions
        for word in shlex.split(dcmd_arg):
            if word in actlevel:
                actlevel = actlevel[word]

        if not actlevel:
            actlevel["SEEN"] = True


def test_dcmd_coverage():
    # This checks all of the dcmd() calls in the AST of this file and
    # matches them up with commands defined in the dcmd command set.
    # Any commands not called will show up in
    # `scanner.coverage_needed()`.

    argparser = argparse.ArgumentParser()
    PcmdCommands._add_to_parser(
        argparser.add_subparsers(title="commands", required=True)
    )
    scanner = CoverageScanner()
    scanner.register_actions(argparser)

    test_ast = ast.parse(open(__file__).read(), filename=__file__)
    scanner.visit(test_ast)

    assert scanner.coverage_needed() == set()



from typing import Any
from typing import Dict
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.data_element_permissible_values_date_time_format_date_time_format_format_type_0_isoformat import (
    DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0Isoformat,
)


T = TypeVar(
    "T", bound="DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0"
)


@_attrs_define
class DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0:
    """DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0 model

    Attributes:
        isoformat (DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0Isoformat): ISO 8601 standardized
            date/time formats
    """

    isoformat: DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0Isoformat

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        isoformat = self.isoformat.value

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "isoformat": isoformat,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0` from a dict"""
        d = src_dict.copy()
        isoformat = DataElementPermissibleValuesDateTimeFormatDateTimeFormatFormatType0Isoformat(
            d.pop("isoformat")
        )

        data_element_permissible_values_date_time_format_date_time_format_format_type_0 = cls(
            isoformat=isoformat,
        )

        return data_element_permissible_values_date_time_format_date_time_format_format_type_0



"""Upload a file, create a link or a folder"""
from http import HTTPStatus
from typing import Any
from typing import Dict
from typing import Union

import httpx

from ... import errors
from ...client import AuthenticatedClient
from ...client import Client
from ...models.collections_items_create_item_files_form_style_upload import (
    CollectionsItemsCreateItemFilesFormStyleUpload,
)
from ...models.collections_items_create_item_json_new_item_request import (
    CollectionsItemsCreateItemJsonNewItemRequest,
)
from ...models.item import Item
from ...models.server_error import ServerError
from ...types import Response


def _get_kwargs(
    collection_id: str,
    version_id: str,
    *,
    body: Union[
        CollectionsItemsCreateItemJsonNewItemRequest,
        CollectionsItemsCreateItemFilesFormStyleUpload,
    ],
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/collections/{collection_id}/{version_id}/items".format(
            collection_id=collection_id,
            version_id=version_id,
        ),
    }

    if isinstance(body, CollectionsItemsCreateItemJsonNewItemRequest):
        _json_body = body.to_dict()

        _kwargs["json"] = _json_body
        headers["Content-Type"] = "application/json"
    if isinstance(body, CollectionsItemsCreateItemFilesFormStyleUpload):
        _files_body = body.to_multipart()

        _kwargs["files"] = _files_body

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[Item, ServerError]:
    if response.status_code == HTTPStatus.CREATED:
        response_201 = Item.from_dict(response.json())

        return response_201
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ServerError.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNAUTHORIZED:
        response_401 = ServerError.from_dict(response.json())

        return response_401
    if response.status_code == HTTPStatus.FORBIDDEN:
        response_403 = ServerError.from_dict(response.json())

        return response_403
    if response.status_code == HTTPStatus.CONFLICT:
        response_409 = ServerError.from_dict(response.json())

        return response_409
    if response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE:
        response_415 = ServerError.from_dict(response.json())

        return response_415
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = ServerError.from_dict(response.json())

        return response_500

    raise errors.UnexpectedStatus(response.status_code, response.content)


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[Item, ServerError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        CollectionsItemsCreateItemJsonNewItemRequest,
        CollectionsItemsCreateItemFilesFormStyleUpload,
    ],
) -> Response[Union[Item, ServerError]]:
    r"""Upload a file, create a link or a folder

    Create a new item (a file, link, folder, or dataview) given its specification.

    Two request body formats are supported:

    * `application/json` is typically used for links, folders, and
      very small files/dataviews. The item content (for files and
      dataviews) can be specified as a string in the request body,
      under the `content.data` field.

    * `multipart/form-data` is used for larger files. The item
      content should be provided in the form `content` field, and
      the item name will be taken as the `content` field's
      `filename` attribute.

    If provided, the `folderId` attribute should either be a
    string of an existing item ID that references a folder in the
    collection, or else the string \"ROOT\", denoting the top level
    of the collection.

    For files and dataviews, the `storage.url` field should
    reference the URL of an existing datastore defined in the
    system (as listed by the `/datastores` API call). It can be
    assumed that the datastore `delfini+datastore://default` will
    always exist, but users may choose to store their data in an
    alternate datastore.

    For links, the `storage.url` field should reference an
    external resource available through a standard URL scheme such
    as `https://` or `s3://`, or another Delfini resource available
    through the URL scheme
    `delfini://hostname/collection-id/version-id/item-id`.

    If the checksum of a file, link, or dataview is known by the
    client at the time of upload, it can be provided in either the
    `storage.checksum` field (supporting all item types) or the
    `content.checksum` field (supporting just files and
    dataviews).

    If the user desires the item to be parseable as a table, the
    `parser` field should be set, with the name of the parser
    provided as well as optional parser options. Otherwise, the
    `parser` field can remain unset.

    If the user knows or wishes to define the column schema of the
    item, a list of column definitions can be provided in the
    `columns` field. This should be omitted if the item will not
    be parsed as a table.

    Args:
        collection_id (str):
        version_id (str):
        body (CollectionsItemsCreateItemJsonNewItemRequest):
        body (CollectionsItemsCreateItemFilesFormStyleUpload): Creation of a new item

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Item, ServerError]]
    """

    kwargs = _get_kwargs(
        collection_id=collection_id,
        version_id=version_id,
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        CollectionsItemsCreateItemJsonNewItemRequest,
        CollectionsItemsCreateItemFilesFormStyleUpload,
    ],
) -> Union[Item]:
    r"""Upload a file, create a link or a folder

    Create a new item (a file, link, folder, or dataview) given its specification.

    Two request body formats are supported:

    * `application/json` is typically used for links, folders, and
      very small files/dataviews. The item content (for files and
      dataviews) can be specified as a string in the request body,
      under the `content.data` field.

    * `multipart/form-data` is used for larger files. The item
      content should be provided in the form `content` field, and
      the item name will be taken as the `content` field's
      `filename` attribute.

    If provided, the `folderId` attribute should either be a
    string of an existing item ID that references a folder in the
    collection, or else the string \"ROOT\", denoting the top level
    of the collection.

    For files and dataviews, the `storage.url` field should
    reference the URL of an existing datastore defined in the
    system (as listed by the `/datastores` API call). It can be
    assumed that the datastore `delfini+datastore://default` will
    always exist, but users may choose to store their data in an
    alternate datastore.

    For links, the `storage.url` field should reference an
    external resource available through a standard URL scheme such
    as `https://` or `s3://`, or another Delfini resource available
    through the URL scheme
    `delfini://hostname/collection-id/version-id/item-id`.

    If the checksum of a file, link, or dataview is known by the
    client at the time of upload, it can be provided in either the
    `storage.checksum` field (supporting all item types) or the
    `content.checksum` field (supporting just files and
    dataviews).

    If the user desires the item to be parseable as a table, the
    `parser` field should be set, with the name of the parser
    provided as well as optional parser options. Otherwise, the
    `parser` field can remain unset.

    If the user knows or wishes to define the column schema of the
    item, a list of column definitions can be provided in the
    `columns` field. This should be omitted if the item will not
    be parsed as a table.

    Args:
        collection_id (str):
        version_id (str):
        body (CollectionsItemsCreateItemJsonNewItemRequest):
        body (CollectionsItemsCreateItemFilesFormStyleUpload): Creation of a new item

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Item]
    """

    response = sync_detailed(
        collection_id=collection_id,
        version_id=version_id,
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed


async def asyncio_detailed(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        CollectionsItemsCreateItemJsonNewItemRequest,
        CollectionsItemsCreateItemFilesFormStyleUpload,
    ],
) -> Response[Union[Item, ServerError]]:
    r"""Upload a file, create a link or a folder

    Create a new item (a file, link, folder, or dataview) given its specification.

    Two request body formats are supported:

    * `application/json` is typically used for links, folders, and
      very small files/dataviews. The item content (for files and
      dataviews) can be specified as a string in the request body,
      under the `content.data` field.

    * `multipart/form-data` is used for larger files. The item
      content should be provided in the form `content` field, and
      the item name will be taken as the `content` field's
      `filename` attribute.

    If provided, the `folderId` attribute should either be a
    string of an existing item ID that references a folder in the
    collection, or else the string \"ROOT\", denoting the top level
    of the collection.

    For files and dataviews, the `storage.url` field should
    reference the URL of an existing datastore defined in the
    system (as listed by the `/datastores` API call). It can be
    assumed that the datastore `delfini+datastore://default` will
    always exist, but users may choose to store their data in an
    alternate datastore.

    For links, the `storage.url` field should reference an
    external resource available through a standard URL scheme such
    as `https://` or `s3://`, or another Delfini resource available
    through the URL scheme
    `delfini://hostname/collection-id/version-id/item-id`.

    If the checksum of a file, link, or dataview is known by the
    client at the time of upload, it can be provided in either the
    `storage.checksum` field (supporting all item types) or the
    `content.checksum` field (supporting just files and
    dataviews).

    If the user desires the item to be parseable as a table, the
    `parser` field should be set, with the name of the parser
    provided as well as optional parser options. Otherwise, the
    `parser` field can remain unset.

    If the user knows or wishes to define the column schema of the
    item, a list of column definitions can be provided in the
    `columns` field. This should be omitted if the item will not
    be parsed as a table.

    Args:
        collection_id (str):
        version_id (str):
        body (CollectionsItemsCreateItemJsonNewItemRequest):
        body (CollectionsItemsCreateItemFilesFormStyleUpload): Creation of a new item

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[Item, ServerError]]
    """

    kwargs = _get_kwargs(
        collection_id=collection_id,
        version_id=version_id,
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    collection_id: str,
    version_id: str,
    *,
    client: AuthenticatedClient,
    body: Union[
        CollectionsItemsCreateItemJsonNewItemRequest,
        CollectionsItemsCreateItemFilesFormStyleUpload,
    ],
) -> Union[Item]:
    r"""Upload a file, create a link or a folder

    Create a new item (a file, link, folder, or dataview) given its specification.

    Two request body formats are supported:

    * `application/json` is typically used for links, folders, and
      very small files/dataviews. The item content (for files and
      dataviews) can be specified as a string in the request body,
      under the `content.data` field.

    * `multipart/form-data` is used for larger files. The item
      content should be provided in the form `content` field, and
      the item name will be taken as the `content` field's
      `filename` attribute.

    If provided, the `folderId` attribute should either be a
    string of an existing item ID that references a folder in the
    collection, or else the string \"ROOT\", denoting the top level
    of the collection.

    For files and dataviews, the `storage.url` field should
    reference the URL of an existing datastore defined in the
    system (as listed by the `/datastores` API call). It can be
    assumed that the datastore `delfini+datastore://default` will
    always exist, but users may choose to store their data in an
    alternate datastore.

    For links, the `storage.url` field should reference an
    external resource available through a standard URL scheme such
    as `https://` or `s3://`, or another Delfini resource available
    through the URL scheme
    `delfini://hostname/collection-id/version-id/item-id`.

    If the checksum of a file, link, or dataview is known by the
    client at the time of upload, it can be provided in either the
    `storage.checksum` field (supporting all item types) or the
    `content.checksum` field (supporting just files and
    dataviews).

    If the user desires the item to be parseable as a table, the
    `parser` field should be set, with the name of the parser
    provided as well as optional parser options. Otherwise, the
    `parser` field can remain unset.

    If the user knows or wishes to define the column schema of the
    item, a list of column definitions can be provided in the
    `columns` field. This should be omitted if the item will not
    be parsed as a table.

    Args:
        collection_id (str):
        version_id (str):
        body (CollectionsItemsCreateItemJsonNewItemRequest):
        body (CollectionsItemsCreateItemFilesFormStyleUpload): Creation of a new item

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[Item]
    """

    response = await asyncio_detailed(
        collection_id=collection_id,
        version_id=version_id,
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed



from enum import Enum


class MetricAggFunc(str, Enum):
    AVG = "avg"
    CARDINALITY = "cardinality"
    COUNT = "count"
    SUM = "sum"

    def __str__(self) -> str:
        return str(self.value)



"""Given a data dictionary, identify candidate item columns that
align and validate
"""
from http import HTTPStatus
from typing import Any
from typing import Dict
from typing import Union

import httpx

from ... import errors
from ...client import AuthenticatedClient
from ...client import Client
from ...models.search_dictionaries_by_item_inverse_response import (
    SearchDictionariesByItemInverseResponse,
)
from ...models.search_search_dictionaries_by_item_inverse_body import (
    SearchSearchDictionariesByItemInverseBody,
)
from ...models.server_error import ServerError
from ...types import Response


def _get_kwargs(
    *,
    body: SearchSearchDictionariesByItemInverseBody,
) -> Dict[str, Any]:
    headers: Dict[str, Any] = {}

    _kwargs: Dict[str, Any] = {
        "method": "post",
        "url": "/search/dictionaries/byiteminverse",
    }

    _body = body.to_dict()

    _kwargs["json"] = _body
    headers["Content-Type"] = "application/json"

    _kwargs["headers"] = headers
    return _kwargs


def _parse_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Union[SearchDictionariesByItemInverseResponse, ServerError]:
    if response.status_code == HTTPStatus.OK:
        response_200 = SearchDictionariesByItemInverseResponse.from_dict(
            response.json()
        )

        return response_200
    if response.status_code == HTTPStatus.BAD_REQUEST:
        response_400 = ServerError.from_dict(response.json())

        return response_400
    if response.status_code == HTTPStatus.UNSUPPORTED_MEDIA_TYPE:
        response_415 = ServerError.from_dict(response.json())

        return response_415
    if response.status_code == HTTPStatus.INTERNAL_SERVER_ERROR:
        response_500 = ServerError.from_dict(response.json())

        return response_500

    raise errors.UnexpectedStatus(response.status_code, response.content)


def _build_response(
    *, client: Union[AuthenticatedClient, Client], response: httpx.Response
) -> Response[Union[SearchDictionariesByItemInverseResponse, ServerError]]:
    return Response(
        status_code=HTTPStatus(response.status_code),
        content=response.content,
        headers=response.headers,
        parsed=_parse_response(client=client, response=response),
    )


def sync_detailed(
    *,
    client: AuthenticatedClient,
    body: SearchSearchDictionariesByItemInverseBody,
) -> Response[Union[SearchDictionariesByItemInverseResponse, ServerError]]:
    """Given a data dictionary, identify candidate item columns that
    align and validate

    This performs a 'reverse search' for columns in the provided
    item that are likely to align with data elements in the
    provided source. The results are returned per data element,
    without pagination currently.

    The current method uses the same approach as the 'forward
    search', performing full-text search queries using the column
    names against the data element definitions. These hits are
    then collated by data element and ranked to provide the most
    likely column hits for each data element.

    Args:
        body (SearchSearchDictionariesByItemInverseBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[SearchDictionariesByItemInverseResponse, ServerError]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = client.get_httpx_client().request(
        **kwargs,
    )

    return _build_response(client=client, response=response)


def sync(
    *,
    client: AuthenticatedClient,
    body: SearchSearchDictionariesByItemInverseBody,
) -> Union[SearchDictionariesByItemInverseResponse]:
    """Given a data dictionary, identify candidate item columns that
    align and validate

    This performs a 'reverse search' for columns in the provided
    item that are likely to align with data elements in the
    provided source. The results are returned per data element,
    without pagination currently.

    The current method uses the same approach as the 'forward
    search', performing full-text search queries using the column
    names against the data element definitions. These hits are
    then collated by data element and ranked to provide the most
    likely column hits for each data element.

    Args:
        body (SearchSearchDictionariesByItemInverseBody):

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[SearchDictionariesByItemInverseResponse]
    """

    response = sync_detailed(
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed


async def asyncio_detailed(
    *,
    client: AuthenticatedClient,
    body: SearchSearchDictionariesByItemInverseBody,
) -> Response[Union[SearchDictionariesByItemInverseResponse, ServerError]]:
    """Given a data dictionary, identify candidate item columns that
    align and validate

    This performs a 'reverse search' for columns in the provided
    item that are likely to align with data elements in the
    provided source. The results are returned per data element,
    without pagination currently.

    The current method uses the same approach as the 'forward
    search', performing full-text search queries using the column
    names against the data element definitions. These hits are
    then collated by data element and ranked to provide the most
    likely column hits for each data element.

    Args:
        body (SearchSearchDictionariesByItemInverseBody):

    Raises:
        errors.UnexpectedStatus: If the server returns an undocumented status code and Client.raise_on_unexpected_status is True.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Response[Union[SearchDictionariesByItemInverseResponse, ServerError]]
    """

    kwargs = _get_kwargs(
        body=body,
    )

    response = await client.get_async_httpx_client().request(**kwargs)

    return _build_response(client=client, response=response)


async def asyncio(
    *,
    client: AuthenticatedClient,
    body: SearchSearchDictionariesByItemInverseBody,
) -> Union[SearchDictionariesByItemInverseResponse]:
    """Given a data dictionary, identify candidate item columns that
    align and validate

    This performs a 'reverse search' for columns in the provided
    item that are likely to align with data elements in the
    provided source. The results are returned per data element,
    without pagination currently.

    The current method uses the same approach as the 'forward
    search', performing full-text search queries using the column
    names against the data element definitions. These hits are
    then collated by data element and ranked to provide the most
    likely column hits for each data element.

    Args:
        body (SearchSearchDictionariesByItemInverseBody):

    Raises:
        errors.UnexpectedStatus: If the server returns a status code greater than or equal to 300.
        httpx.TimeoutException: If the request takes longer than Client.timeout.

    Returns:
        Union[SearchDictionariesByItemInverseResponse]
    """

    response = await asyncio_detailed(
        client=client,
        body=body,
    )
    if isinstance(response.parsed, ServerError):
        raise errors.UnexpectedStatus(response.status_code, response.content)

    return response.parsed



import datetime
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="FederationAddressOrObjectType2")


@_attrs_define
class FederationAddressOrObjectType2:
    """FederationAddressOrObjectType2 model

    Attributes:
        content (Union[Unset, str]):
        name (Union[Unset, str]):
        published (Union[Unset, datetime.datetime]):
        type (Union[Unset, str]):
    """

    content: Union[Unset, str] = UNSET
    name: Union[Unset, str] = UNSET
    published: Union[Unset, datetime.datetime] = UNSET
    type: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        content = self.content
        name = self.name
        published: Union[Unset, str] = UNSET
        if not isinstance(self.published, Unset):
            published = self.published.isoformat()
        type = self.type

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update({})
        if content is not UNSET:
            field_dict["content"] = content
        if name is not UNSET:
            field_dict["name"] = name
        if published is not UNSET:
            field_dict["published"] = published
        if type is not UNSET:
            field_dict["type"] = type

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`FederationAddressOrObjectType2` from a dict"""
        d = src_dict.copy()
        content = d.pop("content", UNSET)

        name = d.pop("name", UNSET)

        _published = d.pop("published", UNSET)
        published: Union[Unset, datetime.datetime]
        if isinstance(_published, Unset):
            published = UNSET
        else:
            published = isoparse(_published)

        type = d.pop("type", UNSET)

        federation_address_or_object_type_2 = cls(
            content=content,
            name=name,
            published=published,
            type=type,
        )

        federation_address_or_object_type_2.additional_properties = d
        return federation_address_or_object_type_2

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.pagination import Pagination
from ..models.search_collections_response_hits_item import (
    SearchCollectionsResponseHitsItem,
)
from ..models.termset import Termset


T = TypeVar("T", bound="SearchCollectionsResponse")


@_attrs_define
class SearchCollectionsResponse:
    """SearchCollectionsResponse model

    Attributes:
        hits (List['SearchCollectionsResponseHitsItem']):
        metadata_terms (Termset):
        pagination (Pagination):
        search_terms (Termset):
    """

    hits: List["SearchCollectionsResponseHitsItem"]
    metadata_terms: "Termset"
    pagination: "Pagination"
    search_terms: "Termset"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        hits = []
        for hits_item_data in self.hits:
            hits_item = hits_item_data.to_dict()
            hits.append(hits_item)

        metadata_terms = self.metadata_terms.to_dict()
        pagination = self.pagination.to_dict()
        search_terms = self.search_terms.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "hits": hits,
                "metadataTerms": metadata_terms,
                "pagination": pagination,
                "searchTerms": search_terms,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SearchCollectionsResponse` from a dict"""
        d = src_dict.copy()
        hits = []
        _hits = d.pop("hits")
        for hits_item_data in _hits:
            hits_item = SearchCollectionsResponseHitsItem.from_dict(hits_item_data)

            hits.append(hits_item)

        metadata_terms = Termset.from_dict(d.pop("metadataTerms"))

        pagination = Pagination.from_dict(d.pop("pagination"))

        search_terms = Termset.from_dict(d.pop("searchTerms"))

        search_collections_response = cls(
            hits=hits,
            metadata_terms=metadata_terms,
            pagination=pagination,
            search_terms=search_terms,
        )

        return search_collections_response



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.collection_cde_stats_item_distrib_item import (
    CollectionCdeStatsItemDistribItem,
)
from ..models.collection_cde_stats_per_cde import CollectionCdeStatsPerCde
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="CollectionCdeStats")


@_attrs_define
class CollectionCdeStats:
    """CollectionCdeStats model

    Attributes:
        per_cde (CollectionCdeStatsPerCde): A map of each CDE url in the requested CDE set to the
            number of items that contain that CDE.
        item_distrib (Union[Unset, List['CollectionCdeStatsItemDistribItem']]): Represents the number of items that have
            a given number of
            CDEs out of the requested set.
    """

    per_cde: "CollectionCdeStatsPerCde"
    item_distrib: Union[Unset, List["CollectionCdeStatsItemDistribItem"]] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        per_cde = self.per_cde.to_dict()
        item_distrib: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.item_distrib, Unset):
            item_distrib = []
            for item_distrib_item_data in self.item_distrib:
                item_distrib_item = item_distrib_item_data.to_dict()
                item_distrib.append(item_distrib_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "perCde": per_cde,
            }
        )
        if item_distrib is not UNSET:
            field_dict["itemDistrib"] = item_distrib

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`CollectionCdeStats` from a dict"""
        d = src_dict.copy()
        per_cde = CollectionCdeStatsPerCde.from_dict(d.pop("perCde"))

        item_distrib = []
        _item_distrib = d.pop("itemDistrib", UNSET)
        for item_distrib_item_data in _item_distrib or []:
            item_distrib_item = CollectionCdeStatsItemDistribItem.from_dict(
                item_distrib_item_data
            )

            item_distrib.append(item_distrib_item)

        collection_cde_stats = cls(
            per_cde=per_cde,
            item_distrib=item_distrib,
        )

        return collection_cde_stats



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.data_element_concept import DataElementConcept
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="DataElementPermissibleValuesValueSetValueSetItem")


@_attrs_define
class DataElementPermissibleValuesValueSetValueSetItem:
    """A single permissible value match.

    Attributes:
        value (Union[None, bool, float, int, str]):
        concepts (Union[Unset, List['DataElementConcept']]): Zero or more concepts associated with this individual
            value. Typically, the `appliesTo` field for PV valueSet concepts will be "valueDomain".
        label (Union[Unset, str]): A short, readable summary of this value.
        meaning (Union[Unset, str]): A description of this value's meaning.
    """

    value: Union[None, bool, float, int, str]
    concepts: Union[Unset, List["DataElementConcept"]] = UNSET
    label: Union[Unset, str] = UNSET
    meaning: Union[Unset, str] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        value: Union[None, bool, float, int, str]
        value = self.value
        concepts: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.concepts, Unset):
            concepts = []
            for concepts_item_data in self.concepts:
                concepts_item = concepts_item_data.to_dict()
                concepts.append(concepts_item)

        label = self.label
        meaning = self.meaning

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "value": value,
            }
        )
        if concepts is not UNSET:
            field_dict["concepts"] = concepts
        if label is not UNSET:
            field_dict["label"] = label
        if meaning is not UNSET:
            field_dict["meaning"] = meaning

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`DataElementPermissibleValuesValueSetValueSetItem` from a dict"""
        d = src_dict.copy()

        def _parse_value(data: object) -> Union[None, bool, float, int, str]:
            if data is None:
                return data
            return cast(Union[None, bool, float, int, str], data)

        value = _parse_value(d.pop("value"))

        concepts = []
        _concepts = d.pop("concepts", UNSET)
        for concepts_item_data in _concepts or []:
            concepts_item = DataElementConcept.from_dict(concepts_item_data)

            concepts.append(concepts_item)

        label = d.pop("label", UNSET)

        meaning = d.pop("meaning", UNSET)

        data_element_permissible_values_value_set_value_set_item = cls(
            value=value,
            concepts=concepts,
            label=label,
            meaning=meaning,
        )

        return data_element_permissible_values_value_set_value_set_item



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.metadata_field import MetadataField
from ..models.metadata_field_group import MetadataFieldGroup
from ..models.metadata_field_groups_data_elements import MetadataFieldGroupsDataElements
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="MetadataFieldGroups")


@_attrs_define
class MetadataFieldGroups:
    """MetadataFieldGroups model

    Attributes:
        fields (List['MetadataField']):
        groups (List['MetadataFieldGroup']):
        data_elements (Union[Unset, MetadataFieldGroupsDataElements]):
    """

    fields: List["MetadataField"]
    groups: List["MetadataFieldGroup"]
    data_elements: Union[Unset, "MetadataFieldGroupsDataElements"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        fields = []
        for fields_item_data in self.fields:
            fields_item = fields_item_data.to_dict()
            fields.append(fields_item)

        groups = []
        for groups_item_data in self.groups:
            groups_item = groups_item_data.to_dict()
            groups.append(groups_item)

        data_elements: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.data_elements, Unset):
            data_elements = self.data_elements.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "fields": fields,
                "groups": groups,
            }
        )
        if data_elements is not UNSET:
            field_dict["data_elements"] = data_elements

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`MetadataFieldGroups` from a dict"""
        d = src_dict.copy()
        fields = []
        _fields = d.pop("fields")
        for fields_item_data in _fields:
            fields_item = MetadataField.from_dict(fields_item_data)

            fields.append(fields_item)

        groups = []
        _groups = d.pop("groups")
        for groups_item_data in _groups:
            groups_item = MetadataFieldGroup.from_dict(groups_item_data)

            groups.append(groups_item)

        _data_elements = d.pop("data_elements", UNSET)
        data_elements: Union[Unset, MetadataFieldGroupsDataElements]
        if isinstance(_data_elements, Unset):
            data_elements = UNSET
        else:
            data_elements = MetadataFieldGroupsDataElements.from_dict(_data_elements)

        metadata_field_groups = cls(
            fields=fields,
            groups=groups,
            data_elements=data_elements,
        )

        return metadata_field_groups



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.new_collection_version_metadata import NewCollectionVersionMetadata
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="NewCollectionVersion")


@_attrs_define
class NewCollectionVersion:
    """Collection version details

    Attributes:
        description (str): The description of the version
        frozen_release (Union[Unset, bool]): If true, the system will store the current state of all
            link targets and dataviews in the default datastore
             Default: False.
        item_ids (Union[Unset, List[str]]): The items to include in this collection version
        metadata (Union[Unset, NewCollectionVersionMetadata]): Arbitrary user-defined metadata, key-value pairs
    """

    description: str
    frozen_release: Union[Unset, bool] = False
    item_ids: Union[Unset, List[str]] = UNSET
    metadata: Union[Unset, "NewCollectionVersionMetadata"] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        description = self.description
        frozen_release = self.frozen_release
        item_ids: Union[Unset, List[str]] = UNSET
        if not isinstance(self.item_ids, Unset):
            item_ids = self.item_ids

        metadata: Union[Unset, Dict[str, Any]] = UNSET
        if not isinstance(self.metadata, Unset):
            metadata = self.metadata.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "description": description,
            }
        )
        if frozen_release is not UNSET:
            field_dict["frozenRelease"] = frozen_release
        if item_ids is not UNSET:
            field_dict["itemIds"] = item_ids
        if metadata is not UNSET:
            field_dict["metadata"] = metadata

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`NewCollectionVersion` from a dict"""
        d = src_dict.copy()
        description = d.pop("description")

        frozen_release = d.pop("frozenRelease", UNSET)

        item_ids = cast(List[str], d.pop("itemIds", UNSET))

        _metadata = d.pop("metadata", UNSET)
        metadata: Union[Unset, NewCollectionVersionMetadata]
        if isinstance(_metadata, Unset):
            metadata = UNSET
        else:
            metadata = NewCollectionVersionMetadata.from_dict(_metadata)

        new_collection_version = cls(
            description=description,
            frozen_release=frozen_release,
            item_ids=item_ids,
            metadata=metadata,
        )

        return new_collection_version



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.metadata_field_group import MetadataFieldGroup


T = TypeVar("T", bound="MetadataSetFieldGroupsBody")


@_attrs_define
class MetadataSetFieldGroupsBody:
    """MetadataSetFieldGroupsBody model

    Attributes:
        groups (List['MetadataFieldGroup']):
    """

    groups: List["MetadataFieldGroup"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        groups = []
        for groups_item_data in self.groups:
            groups_item = groups_item_data.to_dict()
            groups.append(groups_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "groups": groups,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`MetadataSetFieldGroupsBody` from a dict"""
        d = src_dict.copy()
        groups = []
        _groups = d.pop("groups")
        for groups_item_data in _groups:
            groups_item = MetadataFieldGroup.from_dict(groups_item_data)

            groups.append(groups_item)

        metadata_set_field_groups_body = cls(
            groups=groups,
        )

        return metadata_set_field_groups_body



from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define

from ..models.search_search_collections_body_types_item import (
    SearchSearchCollectionsBodyTypesItem,
)
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="SearchSearchCollectionsBody")


@_attrs_define
class SearchSearchCollectionsBody:
    """SearchSearchCollectionsBody model

    Attributes:
        query (str):
        bmeta (Union[Unset, str]): Filter by collection metadata.

            Provide a base64-encoded JSON mapping conformant to the
            following JSON schema:

              schema:
                type: object
                additionalProperties:
                  anyOf:
                    - type: string
                    - type: array
                      items:
                        type: string

            This includes any combination of `{"x-meta": "foo"}` and
            `{"x-meta-2": ["foo", "bar"]}`:

            * In the single string case, the provided string must
              match exactly to the value in the metadata field.

            * In the case of an array of strings, the metadata field
              is assumed to contain a JSON-encoded array of strings,
              and only one of the strings in the provided array needs
              to match one of the strings in the metadata field.
        data_elements (Union[Unset, List[str]]):
        types (Union[Unset, List[SearchSearchCollectionsBodyTypesItem]]):
    """

    query: str
    bmeta: Union[Unset, str] = UNSET
    data_elements: Union[Unset, List[str]] = UNSET
    types: Union[Unset, List[SearchSearchCollectionsBodyTypesItem]] = UNSET

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        query = self.query
        bmeta = self.bmeta
        data_elements: Union[Unset, List[str]] = UNSET
        if not isinstance(self.data_elements, Unset):
            data_elements = self.data_elements

        types: Union[Unset, List[str]] = UNSET
        if not isinstance(self.types, Unset):
            types = []
            for types_item_data in self.types:
                types_item = types_item_data.value
                types.append(types_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "query": query,
            }
        )
        if bmeta is not UNSET:
            field_dict["bmeta"] = bmeta
        if data_elements is not UNSET:
            field_dict["data_elements"] = data_elements
        if types is not UNSET:
            field_dict["types"] = types

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`SearchSearchCollectionsBody` from a dict"""
        d = src_dict.copy()
        query = d.pop("query")

        bmeta = d.pop("bmeta", UNSET)

        data_elements = cast(List[str], d.pop("data_elements", UNSET))

        types = []
        _types = d.pop("types", UNSET)
        for types_item_data in _types or []:
            types_item = SearchSearchCollectionsBodyTypesItem(types_item_data)

            types.append(types_item)

        search_search_collections_body = cls(
            query=query,
            bmeta=bmeta,
            data_elements=data_elements,
            types=types,
        )

        return search_search_collections_body



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.group import Group
from ..models.group_get_groups_response_200_self_roles import (
    GroupGetGroupsResponse200SelfRoles,
)
from ..models.pagination import Pagination


T = TypeVar("T", bound="GroupGetGroupsResponse200")


@_attrs_define
class GroupGetGroupsResponse200:
    """GroupGetGroupsResponse200 model

    Attributes:
        groups (List['Group']):
        pagination (Pagination):
        self_roles (GroupGetGroupsResponse200SelfRoles):
    """

    groups: List["Group"]
    pagination: "Pagination"
    self_roles: "GroupGetGroupsResponse200SelfRoles"

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        groups = []
        for groups_item_data in self.groups:
            groups_item = groups_item_data.to_dict()
            groups.append(groups_item)

        pagination = self.pagination.to_dict()
        self_roles = self.self_roles.to_dict()

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "groups": groups,
                "pagination": pagination,
                "selfRoles": self_roles,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`GroupGetGroupsResponse200` from a dict"""
        d = src_dict.copy()
        groups = []
        _groups = d.pop("groups")
        for groups_item_data in _groups:
            groups_item = Group.from_dict(groups_item_data)

            groups.append(groups_item)

        pagination = Pagination.from_dict(d.pop("pagination"))

        self_roles = GroupGetGroupsResponse200SelfRoles.from_dict(d.pop("selfRoles"))

        group_get_groups_response_200 = cls(
            groups=groups,
            pagination=pagination,
            self_roles=self_roles,
        )

        return group_get_groups_response_200



from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar

from attrs import define as _attrs_define

from ..models.parser import Parser


T = TypeVar("T", bound="ParsersGetParsersResponse200")


@_attrs_define
class ParsersGetParsersResponse200:
    """ParsersGetParsersResponse200 model

    Attributes:
        parsers (List['Parser']):
    """

    parsers: List["Parser"]

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        parsers = []
        for parsers_item_data in self.parsers:
            parsers_item = parsers_item_data.to_dict()
            parsers.append(parsers_item)

        field_dict: Dict[str, Any] = {}
        field_dict.update(
            {
                "parsers": parsers,
            }
        )

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`ParsersGetParsersResponse200` from a dict"""
        d = src_dict.copy()
        parsers = []
        _parsers = d.pop("parsers")
        for parsers_item_data in _parsers:
            parsers_item = Parser.from_dict(parsers_item_data)

            parsers.append(parsers_item)

        parsers_get_parsers_response_200 = cls(
            parsers=parsers,
        )

        return parsers_get_parsers_response_200



import datetime
from typing import Any
from typing import cast
from typing import Dict
from typing import List
from typing import Type
from typing import TypeVar
from typing import Union

from attrs import define as _attrs_define
from attrs import field as _attrs_field
from dateutil.parser import isoparse

from ..models.task_result_data_type_4 import TaskResultDataType4
from ..models.task_result_errors_item import TaskResultErrorsItem
from ..models.task_result_status import TaskResultStatus
from ..types import UNSET
from ..types import Unset


T = TypeVar("T", bound="TaskResult")


@_attrs_define
class TaskResult:
    """TaskResult model

    Attributes:
        event_time (datetime.datetime):
        start_time (datetime.datetime):
        status (TaskResultStatus):
        task_id (str):
        task_idemkey (str):
        worker (str):
        data (Union['TaskResultDataType4', List[Any], Unset, float, int, str]):
        errors (Union[Unset, List['TaskResultErrorsItem']]):
        id (Union[Unset, str]):
    """

    event_time: datetime.datetime
    start_time: datetime.datetime
    status: TaskResultStatus
    task_id: str
    task_idemkey: str
    worker: str
    data: Union["TaskResultDataType4", List[Any], Unset, float, int, str] = UNSET
    errors: Union[Unset, List["TaskResultErrorsItem"]] = UNSET
    id: Union[Unset, str] = UNSET
    additional_properties: Dict[str, Any] = _attrs_field(init=False, factory=dict)

    def to_dict(self) -> Dict[str, Any]:
        """Convert to a dict"""
        event_time = self.event_time.isoformat()
        start_time = self.start_time.isoformat()
        status = self.status.value
        task_id = self.task_id
        task_idemkey = self.task_idemkey
        worker = self.worker
        data: Union[Dict[str, Any], List[Any], Unset, float, int, str]
        if isinstance(self.data, Unset):
            data = UNSET
        elif isinstance(self.data, list):
            data = self.data

        elif isinstance(self.data, TaskResultDataType4):
            data = self.data.to_dict()
        else:
            data = self.data
        errors: Union[Unset, List[Dict[str, Any]]] = UNSET
        if not isinstance(self.errors, Unset):
            errors = []
            for errors_item_data in self.errors:
                errors_item = errors_item_data.to_dict()
                errors.append(errors_item)

        id = self.id

        field_dict: Dict[str, Any] = {}
        field_dict.update(self.additional_properties)
        field_dict.update(
            {
                "event_time": event_time,
                "start_time": start_time,
                "status": status,
                "task_id": task_id,
                "task_idemkey": task_idemkey,
                "worker": worker,
            }
        )
        if data is not UNSET:
            field_dict["data"] = data
        if errors is not UNSET:
            field_dict["errors"] = errors
        if id is not UNSET:
            field_dict["id"] = id

        return field_dict

    @classmethod
    def from_dict(cls: Type[T], src_dict: Dict[str, Any]) -> T:
        """Create an instance of :py:class:`TaskResult` from a dict"""
        d = src_dict.copy()
        event_time = isoparse(d.pop("event_time"))

        start_time = isoparse(d.pop("start_time"))

        status = TaskResultStatus(d.pop("status"))

        task_id = d.pop("task_id")

        task_idemkey = d.pop("task_idemkey")

        worker = d.pop("worker")

        def _parse_data(
            data: object,
        ) -> Union["TaskResultDataType4", List[Any], Unset, float, int, str]:
            if isinstance(data, Unset):
                return data
            try:
                if not isinstance(data, list):
                    raise TypeError()
                data_type_3 = cast(List[Any], data)

                return data_type_3
            except:  # noqa: E722
                pass
            try:
                if not isinstance(data, dict):
                    raise TypeError()
                data_type_4 = TaskResultDataType4.from_dict(data)

                return data_type_4
            except:  # noqa: E722
                pass
            return cast(
                Union["TaskResultDataType4", List[Any], Unset, float, int, str], data
            )

        data = _parse_data(d.pop("data", UNSET))

        errors = []
        _errors = d.pop("errors", UNSET)
        for errors_item_data in _errors or []:
            errors_item = TaskResultErrorsItem.from_dict(errors_item_data)

            errors.append(errors_item)

        id = d.pop("id", UNSET)

        task_result = cls(
            event_time=event_time,
            start_time=start_time,
            status=status,
            task_id=task_id,
            task_idemkey=task_idemkey,
            worker=worker,
            data=data,
            errors=errors,
            id=id,
        )

        task_result.additional_properties = d
        return task_result

    @property
    def additional_keys(self) -> List[str]:
        return list(self.additional_properties.keys())

    def __getitem__(self, key: str) -> Any:
        return self.additional_properties[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self.additional_properties[key] = value

    def __delitem__(self, key: str) -> None:
        del self.additional_properties[key]

    def __contains__(self, key: str) -> bool:
        return key in self.additional_properties



"""Contains a helper to assist with handling paginated responses"""
from collections.abc import Iterator
from types import ModuleType
from typing import Any
from typing import Generic
from typing import get_args
from typing import get_type_hints
from typing import Protocol
from typing import TypeVar
from typing import Union

from .client import AuthenticatedClient
from .client import Client
from .errors import UnexpectedStatus
from .models.pagination import Pagination


class HasPagination(Protocol):
    """A :py:mod:`~typing.Protocol` that matches any model that
    supports :py:mod:`~pydelfini.delfini_core.models.pagination.Pagination`.

    """

    pagination: Pagination


T = TypeVar("T", bound=HasPagination)


class Paginator(Generic[T]):
    """A utility that helps with dealing with paginated API responses.

    Paginated API responses have the property ``pagination`` of type
    :py:mod:`~pydelfini.delfini_core.models.pagination.Pagination`. This
    model has an attribute
    :py:attr:`~pydelfini.delfini_core.models.pagination.Pagination.next_page_url`
    which holds the URL of the next response page; since the
    :py:mod:`~pydelfini.delfini_core` library does not allow you to
    directly interact with URLs, this helper assists with making these
    additional page requests.

    This class takes as arguments the module containing the desired
    API endpoint and the appropriate
    :py:mod:`~pydelfini.delfini_core.client`. Since this is a generic
    class, it is also encouraged for type safety to specify the
    expected return type in brackets. The actual API request is done
    by calling the :py:meth:`paginate` method.

    Example usage::

      from pydelfini.delfini_core.paginator import Paginator
      from pydelfini.delfini_core.api.collections import collections_get_collections
      from pydelfini.delfini_core.models import CollectionsGetCollectionsCollectionList

      pager = Paginator[CollectionsGetCollectionsCollectionList](
          collections_get_collections, client
      )
      for page in pager.paginate(public=True):
          # Note that you still need to iterate through each item
          # in each page
          for collection in page.collections:
              print(collection.name)

    .. warning::

      While the return type of :py:meth:`paginate` is set according to
      the generic class definition, the static type checker cannot
      verify that the provided module's :py:func:`sync` method
      actually returns the provided type. Additionally, the arguments
      to the :py:meth:`paginate` method are not type checked to align
      with the :py:func:`sync` method.

    """

    def __init__(
        self,
        mod: ModuleType,
        client: Union[Client, AuthenticatedClient],
    ) -> None:
        self.mod = mod
        self.client = client
        self.response_type = get_type_hints(mod.sync)["return"]

        # NOTE: this check is currently disabled due to the theory
        # that generic type annotations are not used at runtime, and
        # given that we can parse the proper response type at runtime
        # from `mod.sync` we don't even need this check. It can be
        # enabled, though, if users want more safety checks, as
        # skipping this check could theoretically cause the return
        # type annotation of paginate to not match what is actually
        # returned by the runtime code.
        if False and hasattr(self, "__orig_class__"):  # type: ignore[unreachable]
            annotated_type = get_args(self.__orig_class__)[0]  # type: ignore[unreachable]
            if self.response_type != annotated_type:
                raise RuntimeError("Mismatched type annotation")

    def paginate(self, *args: Any, **kwargs: Any) -> Iterator[T]:
        """Get an iterator over all pages from the response.

        Args:
            *args:
                Positional arguments to be passed to the module's :py:func:`sync`
                function
            **kwargs:
                Keyword arguments to be passed to the module's :py:func:`sync`
                function.

        Returns:
            An iterator over each response page.

        """
        parsed = self.mod.sync(*args, client=self.client, **kwargs)
        assert isinstance(parsed, self.response_type)
        yield parsed

        while parsed.pagination.next_page_url:
            httpx_response = self.client.get_httpx_client().request(
                method="get",
                url=parsed.pagination.next_page_url,
            )
            response = self.mod._build_response(
                client=self.client, response=httpx_response
            )
            parsed = response.parsed

            if isinstance(parsed, self.response_type):
                yield parsed
            else:
                raise UnexpectedStatus(response.status_code, response.content)
